"""
    Test cases for salt.modules.win_psget
"""

import pytest

import salt.modules.win_psget as win_psget
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_psget: {}}


@pytest.fixture
def bootstrap_ps_str():
    return """<?xml version="1.0" encoding="utf-8"?>
<Objects>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">NuGet</Property>
    <Property Name="Version" Type="Microsoft.PackageManagement.Internal.Utility.Versions.FourPartVersion">
      <Property Name="Major" Type="System.UInt16">2</Property>
      <Property Name="Minor" Type="System.UInt16">8</Property>
      <Property Name="Build" Type="System.UInt16">5</Property>
      <Property Name="Revision" Type="System.UInt16">208</Property>
    </Property>
    <Property Name="ProviderPath" Type="System.String">C:\\Program Files\\PackageManagement\\ProviderAssemblies\\nuget\\2.8.5
.208\\Microsoft.PackageManagement.NuGetProvider.dll</Property>
  </Object>
</Objects>"""


@pytest.fixture
def avail_modules_ps_str():
    return """<?xml version="1.0" encoding="utf-8"?>
<Objects>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">ActOnCmdlets</Property>
    <Property Name="Description" Type="System.String">CData Cmdlets for Act-On</Property>
  </Object>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">FinancialEdgeNXTCmdlets</Property>
    <Property Name="Description" Type="System.String">CData Cmdlets for Blackbaud Financial Edge NXT</Property>
  </Object>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">GoogleCMCmdlets</Property>
    <Property Name="Description" Type="System.String">CData Cmdlets for Google Campaign Manager</Property>
  </Object>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">DHCPMigration</Property>
    <Property Name="Description" Type="System.String">A module to copy various DHCP information from 1 server to another.</Property>
  </Object>
</Objects>"""


def test_bootstrap(bootstrap_ps_str):
    mock_read_ok = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": bootstrap_ps_str,
        }
    )

    with patch.dict(win_psget.__salt__, {"cmd.run_all": mock_read_ok}):
        assert "NuGet" in win_psget.bootstrap()


def test_avail_modules(avail_modules_ps_str):
    mock_read_ok = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": avail_modules_ps_str,
        }
    )

    with patch.dict(win_psget.__salt__, {"cmd.run_all": mock_read_ok}):
        assert "DHCPMigration" in win_psget.avail_modules(False)
        assert "DHCPMigration" in win_psget.avail_modules(True)



"""
Homebrew for macOS

.. important::
    If you feel that Salt should be using this module to manage packages on a
    minion, and it is using a different module (or gives an error similar to
    *'pkg.install' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import copy
import logging

import salt.utils.data
import salt.utils.functools
import salt.utils.json
import salt.utils.path
import salt.utils.pkg
import salt.utils.versions
from salt.exceptions import CommandExecutionError, MinionError, SaltInvocationError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Confine this module to Mac OS with Homebrew.
    """
    if __grains__["os"] != "MacOS":
        return False, "brew module is macos specific"
    if not salt.utils.path.which("brew"):
        return False, "The 'brew' binary was not found"
    return __virtualname__


def _list_taps():
    """
    List currently installed brew taps
    """
    return _call_brew("tap")["stdout"].splitlines()


def _list_pinned():
    """
    List currently pinned formulas
    """
    return _call_brew("list", "--pinned")["stdout"].splitlines()


def _pin(pkg, runas=None):
    """
    Pin pkg
    """
    try:
        _call_brew("pin", pkg)
    except CommandExecutionError:
        log.error('Failed to pin "%s"', pkg)
        return False

    return True


def _unpin(pkg, runas=None):
    """
    Pin pkg
    """
    try:
        _call_brew("unpin", pkg)
    except CommandExecutionError:
        log.error('Failed to unpin "%s"', pkg)
        return False

    return True


def _tap(tap, runas=None):
    """
    Add unofficial GitHub repos to the list of formulas that brew tracks,
    updates, and installs from.
    """
    if tap in _list_taps():
        return True

    try:
        _call_brew("tap", tap)
    except CommandExecutionError:
        log.error('Failed to tap "%s"', tap)
        return False

    return True


def _homebrew_bin():
    """
    Returns the full path to the homebrew binary in the PATH
    """
    ret = __salt__["cmd.run"]("brew --prefix", output_loglevel="trace")
    ret += "/bin/brew"
    return ret


def _call_brew(*cmd, failhard=True):
    """
    Calls the brew command with the user account of brew
    """
    user = __salt__["file.get_user"](_homebrew_bin())
    runas = user if user != __opts__["user"] else None
    _cmd = []
    if runas:
        _cmd = [f"sudo -i -n -H -u {runas} -- "]
    _cmd = _cmd + [salt.utils.path.which("brew")] + list(cmd)
    _cmd = " ".join(_cmd)

    runas = None
    result = __salt__["cmd.run_all"](
        cmd=_cmd,
        runas=runas,
        output_loglevel="trace",
        python_shell=False,
    )
    if failhard and result["retcode"] != 0:
        raise CommandExecutionError("Brew command failed", info={"result": result})
    return result


def _list_pkgs_from_context(versions_as_list):
    """
    Use pkg list from __context__
    """
    if versions_as_list:
        return __context__["pkg.list_pkgs"]
    else:
        ret = copy.deepcopy(__context__["pkg.list_pkgs"])
        __salt__["pkg_resource.stringify"](ret)
        return ret


def list_pkgs(versions_as_list=False, **kwargs):
    """
    List the packages currently installed in a dict::

        {'<package_name>': '<version>'}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_pkgs
    """
    versions_as_list = salt.utils.data.is_true(versions_as_list)
    # not yet implemented or not applicable
    if any(
        [salt.utils.data.is_true(kwargs.get(x)) for x in ("removed", "purge_desired")]
    ):
        return {}

    if "pkg.list_pkgs" in __context__ and kwargs.get("use_context", True):
        return _list_pkgs_from_context(versions_as_list)

    ret = {}
    package_info = salt.utils.json.loads(
        _call_brew("info", "--json=v2", "--installed")["stdout"]
    )

    for package in package_info["formulae"]:
        # Brew allows multiple versions of the same package to be installed.
        # Salt allows for this, so it must be accounted for.
        pkg_versions = [v["version"] for v in package["installed"]]
        # Brew allows for aliasing of packages, all of which will be
        # installable from a Salt call, so all names must be accounted for.
        pkg_names = package["aliases"] + [package["name"], package["full_name"]]
        # Create a list of tuples containing all possible combinations of
        # names and versions, because all are valid.
        combinations = [(n, v) for n in pkg_names for v in pkg_versions]

        for pkg_name, pkg_version in combinations:
            __salt__["pkg_resource.add_pkg"](ret, pkg_name, pkg_version)

    for package in package_info["casks"]:
        pkg_version = package["installed"]
        pkg_names = {package["full_token"], package["token"]}
        pkg_tap = package.get("tap", None)
        # The following name is appended to maintain backward compatibility
        # with old salt formulas. Since full_token and token are the same
        # for official taps (homebrew/*).
        if not pkg_tap:
            # Tap is null when the package is from homebrew/cask.
            pkg_tap = "homebrew/cask"
        pkg_names.add("/".join([pkg_tap, package["token"]]))
        for pkg_name in pkg_names:
            __salt__["pkg_resource.add_pkg"](ret, pkg_name, pkg_version)

    __salt__["pkg_resource.sort_pkglist"](ret)
    __context__["pkg.list_pkgs"] = copy.deepcopy(ret)
    if not versions_as_list:
        __salt__["pkg_resource.stringify"](ret)
    return ret


def version(*names, **kwargs):
    """
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3>
    """
    return __salt__["pkg_resource.version"](*names, **kwargs)


def latest_version(*names, **kwargs):
    """
    Return the latest version of the named package available for upgrade or
    installation

    Currently chooses stable versions, falling back to devel if that does not
    exist.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
        salt '*' pkg.latest_version <package1> <package2> <package3>
    """
    refresh = salt.utils.data.is_true(kwargs.pop("refresh", True))
    if refresh:
        refresh_db()

    def get_version(pkg_info):
        # Perhaps this will need an option to pick devel by default
        version = pkg_info["versions"]["stable"] or pkg_info["versions"]["devel"]
        if pkg_info["versions"]["bottle"] and pkg_info["revision"] >= 1:
            version = "{}_{}".format(version, pkg_info["revision"])
        return version

    versions_dict = {key: get_version(val) for key, val in _info(*names).items()}

    if len(names) == 1:
        return next(iter(versions_dict.values()))
    else:
        return versions_dict


# available_version is being deprecated
available_version = salt.utils.functools.alias_function(
    latest_version, "available_version"
)


def remove(name=None, pkgs=None, **kwargs):
    """
    Removes packages with ``brew uninstall``.

    name
        The name of the package to be deleted.


    Multiple Package Options:

    pkgs
        A list of packages to delete. Must be passed as a python list. The
        ``name`` parameter will be ignored if this option is passed.

    .. versionadded:: 0.16.0


    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.remove <package name>
        salt '*' pkg.remove <package1>,<package2>,<package3>
        salt '*' pkg.remove pkgs='["foo", "bar"]'
    """
    try:
        pkg_params = __salt__["pkg_resource.parse_targets"](name, pkgs, **kwargs)[0]
    except MinionError as exc:
        raise CommandExecutionError(exc)

    old = list_pkgs()
    targets = [x for x in pkg_params if x in old]
    if not targets:
        return {}

    out = _call_brew("uninstall", *targets)
    if out["retcode"] != 0 and out["stderr"]:
        errors = [out["stderr"]]
    else:
        errors = []

    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            "Problem encountered removing package(s)",
            info={"errors": errors, "changes": ret},
        )

    return ret


def refresh_db(**kwargs):
    """
    Update the homebrew package repository.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.refresh_db
    """
    # Remove rtag file to keep multiple refreshes from happening in pkg states
    salt.utils.pkg.clear_rtag(__opts__)
    if _call_brew("update")["retcode"]:
        log.error("Failed to update")
        return False

    return True


def _info(*pkgs):
    """
    Get all info brew can provide about a list of packages.

    Does not do any kind of processing, so the format depends entirely on
    the output brew gives. This may change if a new version of the format is
    requested.

    On failure, returns an empty dict and logs failure.
    On success, returns a dict mapping each item in pkgs to its corresponding
    object in the output of 'brew info'.

    Caveat: If one of the packages does not exist, no packages will be
            included in the output.
    """
    brew_result = _call_brew("info", "--json=v2", *pkgs)
    if brew_result["retcode"]:
        log.error("Failed to get info about packages: %s", " ".join(pkgs))
        return {}
    output = salt.utils.json.loads(brew_result["stdout"])

    meta_info = {"formulae": ["name", "full_name"], "casks": ["token", "full_token"]}

    pkgs_info = dict()
    for tap, keys in meta_info.items():
        data = output[tap]
        if len(data) == 0:
            continue

        for _pkg in data:
            for key in keys:
                if _pkg[key] in pkgs:
                    pkgs_info[_pkg[key]] = _pkg

    return pkgs_info


def install(name=None, pkgs=None, taps=None, options=None, **kwargs):
    """
    Install the passed package(s) with ``brew install``

    name
        The name of the formula to be installed. Note that this parameter is
        ignored if "pkgs" is passed.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install <package name>

    taps
        Unofficial GitHub repos to use when updating and installing formulas.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install <package name> tap='<tap>'
            salt '*' pkg.install zlib taps='homebrew/dupes'
            salt '*' pkg.install php54 taps='["josegonzalez/php", "homebrew/dupes"]'

    options
        Options to pass to brew. Only applies to initial install. Due to how brew
        works, modifying chosen options requires a full uninstall followed by a
        fresh install. Note that if "pkgs" is used, all options will be passed
        to all packages. Unrecognized options for a package will be silently
        ignored by brew.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install <package name> tap='<tap>'
            salt '*' pkg.install php54 taps='["josegonzalez/php", "homebrew/dupes"]' options='["--with-fpm"]'

    Multiple Package Installation Options:

    pkgs
        A list of formulas to install. Must be passed as a python list.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install pkgs='["foo","bar"]'


    Returns a dict containing the new package names and versions::

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.install 'package package package'
    """
    try:
        pkg_params, pkg_type = __salt__["pkg_resource.parse_targets"](
            name, pkgs, kwargs.get("sources", {})
        )
    except MinionError as exc:
        raise CommandExecutionError(exc)

    if not pkg_params:
        return {}

    cmd = ["install"]
    cmd.extend(list(pkg_params))

    old = list_pkgs()

    # Ensure we've tapped the repo if necessary
    if taps:
        if not isinstance(taps, list):
            # Feels like there is a better way to allow for tap being
            # specified as both a string and a list
            taps = [taps]

        for tap in taps:
            _tap(tap)

    if options:
        cmd.extend(options)

    out = _call_brew(*cmd)
    if out["retcode"] != 0 and out["stderr"]:
        errors = [out["stderr"]]
    else:
        errors = []

    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            "Problem encountered installing package(s)",
            info={"errors": errors, "changes": ret},
        )

    return ret


def list_upgrades(refresh=True, include_casks=False, **kwargs):  # pylint: disable=W0613
    """
    Check whether or not an upgrade is available for all packages

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_upgrades
    """
    if refresh:
        refresh_db()

    res = _call_brew("outdated", "--json=v2")
    ret = {}

    try:
        data = salt.utils.json.loads(res["stdout"])
    except ValueError as err:
        msg = f'unable to interpret output from "brew outdated": {err}'
        log.error(msg)
        raise CommandExecutionError(msg)

    for pkg in data["formulae"]:
        # current means latest available to brew
        ret[pkg["name"]] = pkg["current_version"]

    if include_casks:
        for pkg in data["casks"]:
            # current means latest available to brew
            ret[pkg["name"]] = pkg["current_version"]

    return ret


def upgrade_available(pkg, **kwargs):
    """
    Check whether or not an upgrade is available for a given package

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade_available <package name>
    """
    return pkg in list_upgrades(**kwargs)


def upgrade(refresh=True, **kwargs):
    """
    Upgrade outdated, unpinned brews.

    refresh
        Fetch the newest version of Homebrew and all formulae from GitHub before installing.

    Returns a dictionary containing the changes:

    .. code-block:: python

        {'<package>':  {'old': '<old-version>',
                        'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade
    """
    ret = {
        "changes": {},
        "result": True,
        "comment": "",
    }

    old = list_pkgs()

    if salt.utils.data.is_true(refresh):
        refresh_db()

    result = _call_brew("upgrade", failhard=False)
    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if result["retcode"] != 0:
        raise CommandExecutionError(
            "Problem encountered upgrading packages",
            info={"changes": ret, "result": result},
        )

    return ret


def info_installed(*names, **kwargs):
    """
    Return the information of the named package(s) installed on the system.

    .. versionadded:: 2016.3.1

    names
        The names of the packages for which to return information.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.info_installed <package1>
        salt '*' pkg.info_installed <package1> <package2> <package3> ...
    """
    return _info(*names)


def hold(name=None, pkgs=None, sources=None, **kwargs):  # pylint: disable=W0613
    """
    Set package in 'hold' state, meaning it will not be upgraded.

    .. versionadded:: 3001

    name
        The name of the package, e.g., 'tmux'

    CLI Example:

     .. code-block:: bash

        salt '*' pkg.hold <package name>

    pkgs
        A list of packages to hold. Must be passed as a python list.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.hold pkgs='["foo", "bar"]'
    """
    if not name and not pkgs and not sources:
        raise SaltInvocationError("One of name, pkgs, or sources must be specified.")
    if pkgs and sources:
        raise SaltInvocationError("Only one of pkgs or sources can be specified.")

    targets = []
    if pkgs:
        targets.extend(pkgs)
    elif sources:
        for source in sources:
            targets.append(next(iter(source)))
    else:
        targets.append(name)

    ret = {}
    pinned = _list_pinned()
    installed = list_pkgs()
    for target in targets:
        if isinstance(target, dict):
            target = next(iter(target))

        ret[target] = {"name": target, "changes": {}, "result": False, "comment": ""}

        if target not in installed:
            ret[target]["comment"] = f"Package {target} does not have a state."
        elif target not in pinned:
            if "test" in __opts__ and __opts__["test"]:
                ret[target].update(result=None)
                ret[target]["comment"] = f"Package {target} is set to be held."
            else:
                result = _pin(target)
                if result:
                    changes = {"old": "install", "new": "hold"}
                    ret[target].update(changes=changes, result=True)
                    ret[target]["comment"] = "Package {} is now being held.".format(
                        target
                    )
                else:
                    ret[target].update(result=False)
                    ret[target]["comment"] = f"Unable to hold package {target}."
        else:
            ret[target].update(result=True)
            ret[target]["comment"] = "Package {} is already set to be held.".format(
                target
            )
    return ret


pin = hold


def unhold(name=None, pkgs=None, sources=None, **kwargs):  # pylint: disable=W0613
    """
    Set package current in 'hold' state to install state,
    meaning it will be upgraded.

    .. versionadded:: 3001

    name
        The name of the package, e.g., 'tmux'

     CLI Example:

     .. code-block:: bash

        salt '*' pkg.unhold <package name>

    pkgs
        A list of packages to unhold. Must be passed as a python list.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.unhold pkgs='["foo", "bar"]'
    """
    if not name and not pkgs and not sources:
        raise SaltInvocationError("One of name, pkgs, or sources must be specified.")
    if pkgs and sources:
        raise SaltInvocationError("Only one of pkgs or sources can be specified.")

    targets = []
    if pkgs:
        targets.extend(pkgs)
    elif sources:
        for source in sources:
            targets.append(next(iter(source)))
    else:
        targets.append(name)

    ret = {}
    pinned = _list_pinned()
    installed = list_pkgs()
    for target in targets:
        if isinstance(target, dict):
            target = next(iter(target))

        ret[target] = {"name": target, "changes": {}, "result": False, "comment": ""}

        if target not in installed:
            ret[target]["comment"] = f"Package {target} does not have a state."
        elif target in pinned:
            if "test" in __opts__ and __opts__["test"]:
                ret[target].update(result=None)
                ret[target]["comment"] = "Package {} is set to be unheld.".format(
                    target
                )
            else:
                result = _unpin(target)
                if result:
                    changes = {"old": "hold", "new": "install"}
                    ret[target].update(changes=changes, result=True)
                    ret[target][
                        "comment"
                    ] = f"Package {target} is no longer being held."
                else:
                    ret[target].update(result=False)
                    ret[target]["comment"] = "Unable to unhold package {}.".format(
                        target
                    )
        else:
            ret[target].update(result=True)
            ret[target]["comment"] = "Package {} is already set not to be held.".format(
                target
            )
    return ret


unpin = unhold



import pytest

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account() as _account:
        yield _account


@pytest.mark.parametrize(
    "exit_code, return_code, result",
    [
        (300, 0, True),
        (299, 299, False),
    ],
)
def test_cmd_exitcode(modules, state_tree, exit_code, return_code, result):
    """
    Test receiving an exit code with cmd.run
    """
    ret = modules.state.single(
        "cmd.run",
        name=f"exit {exit_code}",
        shell="cmd",
        success_retcodes=[2, 44, 300],
    )
    assert ret.result is result
    assert ret.filtered["changes"]["retcode"] == return_code


@pytest.mark.parametrize(
    "exit_code, return_code, result",
    [
        (300, 0, True),
        (299, 299, False),
    ],
)
def test_cmd_exitcode_runas(
    modules, state_tree, exit_code, return_code, result, account
):
    ret = modules.state.single(
        "cmd.run",
        name=f"exit {exit_code}",
        shell="cmd",
        success_retcodes=[2, 44, 300],
        runas=account.username,
        password=account.password,
    )
    assert ret.result is result
    assert ret.filtered["changes"]["retcode"] == return_code


@pytest.mark.parametrize(
    "command, expected",
    [
        ("echo foo", "foo"),
        ("cmd /c echo foo", "foo"),
        ("whoami && echo foo", "foo"),
        ("echo \"foo 'bar'\"", "\"foo 'bar'\""),
        ('echo|set /p="foo" & echo|set /p="bar"', "foobar"),
        ('''echo "&<>[]|{}^=;!'+,`~ "''', '''"&<>[]|{}^=;!'+,`~ "'''),
    ],
)
def test_cmd_builtins(modules, command, expected):
    """
    Test builtin cmd.exe commands
    """
    # Final test is failing... can't get it to pass
    result = modules.cmd.run(command, python_shell=True)
    assert expected in result


@pytest.mark.parametrize(
    "command, expected",
    [
        ("echo foo", "foo"),
        ("cmd /c echo foo", "foo"),
        ("whoami && echo foo", "foo"),
        ("echo \"foo 'bar'\"", "\"foo 'bar'\""),
        ('echo|set /p="foo" & echo|set /p="bar"', "foobar"),
        ('''echo "&<>[]|{}^=;!'+,`~ "''', '''"&<>[]|{}^=;!'+,`~ "'''),
    ],
)
def test_cmd_builtins_runas(modules, account, command, expected):
    """
    Test builtin cmd.exe commands with runas
    """
    result = modules.cmd.run(
        cmd=command,
        python_shell=True,
        runas=account.username,
        password=account.password,
    )
    assert expected in result


@pytest.mark.parametrize(
    "command, expected",
    [
        (["whoami.exe", "/?"], 0),
        (["whoami.exe", "/foo"], 1),
        ("whoami.exe /?", 0),
        ("whoami.exe /foo", 1),
    ],
)
def test_binary(modules, command, expected):
    """
    Test running a binary with cmd.run_all
    """
    result = modules.cmd.run_all(cmd=command)
    assert isinstance(result["pid"], int)
    assert result["retcode"] == expected


@pytest.mark.parametrize(
    "command, expected",
    [
        (["whoami.exe", "/?"], 0),
        (["whoami.exe", "/foo"], 1),
        ("whoami.exe /?", 0),
        ("whoami.exe /foo", 1),
    ],
)
def test_binary_runas(modules, account, command, expected):
    """
    Test running a binary with cmd.run_all and runas
    """
    result = modules.cmd.run_all(
        cmd=command, runas=account.username, password=account.password
    )
    assert isinstance(result["pid"], int)
    assert result["retcode"] == expected


@pytest.mark.parametrize(
    "command, env, expected",
    [
        ("echo %a%%b%", {"a": "foo", "b": "bar"}, "foobar"),
    ],
)
def test_cmd_env(modules, command, env, expected):
    """
    Test cmd.run with environment variables
    """
    result = modules.cmd.run_all(command, python_shell=True, env=env)
    assert isinstance(result["pid"], int)
    assert result["retcode"] == 0
    assert result["stdout"] == expected
    assert result["stderr"] == ""


@pytest.mark.parametrize(
    "command, expected, redirect_stderr",
    [
        (["whoami.exe", "/foo"], "/foo", True),
        (["whoami.exe", "/foo"], "/foo", False),
    ],
)
def test_redirect_stderr(modules, command, expected, redirect_stderr):
    """
    Test redirection of stderr to stdout by running cmd.run_all with invalid commands
    """
    result = modules.cmd.run_all(command, redirect_stderr=redirect_stderr)
    assert isinstance(result["pid"], int)
    assert result["retcode"] == 1
    if redirect_stderr:
        assert expected in result["stdout"]
        assert result["stderr"] == ""
    else:
        assert result["stdout"] == ""
        assert expected in result["stderr"]



import logging
import os
import subprocess
import sys

import pytest

import salt

log = logging.getLogger(__name__)


def test_tornado_import_override(tmp_path):
    """
    Ensure we are not using any non vendor'ed tornado
    """
    test_source = """
    from __future__ import absolute_import, print_function
    import salt
    import tornado
    print(tornado.__name__)
    """
    tornado_source = """
    foo = 'bar'
    """
    with pytest.helpers.temp_file(
        "test.py", directory=tmp_path, contents=test_source
    ) as test_source_path, pytest.helpers.temp_file(
        "tornado.py", directory=tmp_path, contents=tornado_source
    ):
        env = os.environ.copy()
        env["PYTHONPATH"] = os.pathsep.join(sys.path)
        ret = subprocess.run(
            [sys.executable, str(test_source_path)],
            capture_output=True,
            env=env,
            shell=False,
            check=False,
            text=True,
        )
        assert ret.returncode == 0
        assert ret.stdout.strip() == "salt.ext.tornado"


def test_regression_56063():
    importer = salt.TornadoImporter()
    try:
        importer.find_module("tornado")
    except TypeError:
        assert False, "TornadoImporter raised type error when one argument passed"



import errno
import logging
import socket
import time

import pytest
import zmq
from pytestshellutils.utils import ports

import salt.utils.stringutils
from salt.log_handlers.logstash_mod import DatagramLogstashHandler, ZMQLogstashHander

log = logging.getLogger(__name__)


@pytest.fixture
def datagram_server():
    logger = logging.getLogger("test_logstash_logger")
    server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    port = ports.get_unused_localhost_port()
    handler = DatagramLogstashHandler("127.0.0.1", port)
    try:
        server.bind(("127.0.0.1", port))
        server.settimeout(2)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        yield server
    finally:
        logger.removeHandler(handler)
        server.close()


@pytest.fixture
def zmq_server():
    logger = logging.getLogger("test_logstash_logger")
    context = zmq.Context()
    server = context.socket(zmq.SUB)
    port = ports.get_unused_localhost_port()
    handler = ZMQLogstashHander(f"tcp://127.0.0.1:{port}")
    try:
        server.setsockopt(zmq.SUBSCRIBE, b"")
        server.bind(f"tcp://127.0.0.1:{port}")

        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        yield server
    finally:
        logger.removeHandler(handler)
        server.close()
        context.term()


@pytest.mark.slow_test
def test_datagram_handler_log_pickling(datagram_server):
    # given
    the_log = "test message"

    # when
    logger = logging.getLogger("test_logstash_logger")
    logger.info(the_log)

    # then
    received_log, _ = datagram_server.recvfrom(12)
    assert received_log == salt.utils.stringutils.to_bytes(the_log)


def test_zmq_handler_log_pickling(zmq_server):
    # given
    the_log = "test message"
    attempts = 5
    received_log = "wrong message"
    logger = logging.getLogger("test_logstash_logger")

    # I couldn't receive the first log message, that's why it is done using loop...
    # https://zeromq.jira.com/browse/LIBZMQ-270 could be related
    while attempts >= 0:
        try:
            # when
            logger.info(the_log)
            time.sleep(0.15)
            received_log = zmq_server.recv(zmq.NOBLOCK)
            # then
            break
        except zmq.ZMQError as exc:
            if exc.errno == errno.EAGAIN:
                attempts -= 1
                time.sleep(0.15)
                continue
            raise

    assert received_log == salt.utils.stringutils.to_bytes(the_log)



"""
    :codeauthor: Bo Maryniuk (bo@suse.de)
    unit.utils.decorators_test
"""

import inspect

import salt.utils.decorators as decorators
from salt.exceptions import (
    CommandExecutionError,
    SaltConfigurationError,
    SaltInvocationError,
)
from salt.version import SaltStackVersion
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class DummyLogger:
    """
    Dummy logger accepts everything and simply logs
    """

    def __init__(self, messages):
        self._messages = messages

    def __getattr__(self, item):
        return self._log

    def _log(self, msg):
        self._messages.append(msg)


class DecoratorsTest(TestCase):
    """
    Testing decorators.
    """

    def old_function(self):
        return "old"

    def new_function(self):
        return "new"

    def _new_function(self):
        return "old"

    def _mk_version(self, name):
        """
        Make a version

        :return:
        """
        return name, SaltStackVersion.from_name(name)

    def arg_function(self, arg1=None, arg2=None, arg3=None):
        return "old"

    def setUp(self):
        """
        Setup a test
        :return:
        """
        self.globs = {
            "__virtualname__": "test",
            "__opts__": {},
            "__pillar__": {},
            "old_function": self.old_function,
            "new_function": self.new_function,
            "_new_function": self._new_function,
        }
        self.addCleanup(delattr, self, "globs")
        self.messages = list()
        self.addCleanup(delattr, self, "messages")
        patcher = patch.object(decorators, "log", DummyLogger(self.messages))
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_is_deprecated_version_eol(self):
        """
        Use of is_deprecated will result to the exception,
        if the expiration version is lower than the current version.
        A successor function is not pointed out.

        :return:
        """
        depr = decorators.is_deprecated(self.globs, "Helium")
        depr._curr_version = self._mk_version("Beryllium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.old_function)()
        self.assertEqual(
            self.messages, ['The lifetime of the function "old_function" expired.']
        )

    def test_is_deprecated_with_successor_eol(self):
        """
        Use of is_deprecated will result to the exception,
        if the expiration version is lower than the current version.
        A successor function is pointed out.

        :return:
        """
        depr = decorators.is_deprecated(
            self.globs, "Helium", with_successor="new_function"
        )
        depr._curr_version = self._mk_version("Beryllium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.old_function)()
        self.assertEqual(
            self.messages,
            [
                'The lifetime of the function "old_function" expired. '
                'Please use its successor "new_function" instead.'
            ],
        )

    def test_is_deprecated(self):
        """
        Use of is_deprecated will result to the log message,
        if the expiration version is higher than the current version.
        A successor function is not pointed out.

        :return:
        """
        depr = decorators.is_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        self.assertEqual(depr(self.old_function)(), self.old_function())
        self.assertEqual(
            self.messages,
            [
                'The function "old_function" is deprecated '
                'and will expire in version "Beryllium".'
            ],
        )

    def test_is_deprecated_with_successor(self):
        """
        Use of is_deprecated will result to the log message,
        if the expiration version is higher than the current version.
        A successor function is pointed out.

        :return:
        """
        depr = decorators.is_deprecated(
            self.globs, "Beryllium", with_successor="old_function"
        )
        depr._curr_version = self._mk_version("Helium")[1]
        self.assertEqual(depr(self.old_function)(), self.old_function())
        self.assertEqual(
            self.messages,
            [
                'The function "old_function" is deprecated '
                'and will expire in version "Beryllium". '
                'Use successor "old_function" instead.'
            ],
        )

    def test_with_deprecated_notfound(self):
        """
        Test with_deprecated should raise an exception, if a same name
        function with the "_" prefix not implemented.

        :return:
        """
        del self.globs["_new_function"]
        self.globs["__opts__"]["use_deprecated"] = ["test.new_function"]
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.new_function)()
        self.assertEqual(
            self.messages,
            [
                'The function "test.new_function" is using its deprecated '
                'version and will expire in version "Beryllium".'
            ],
        )

    def test_with_deprecated_notfound_in_pillar(self):
        """
        Test with_deprecated should raise an exception, if a same name
        function with the "_" prefix not implemented.

        :return:
        """
        del self.globs["_new_function"]
        self.globs["__pillar__"]["use_deprecated"] = ["test.new_function"]
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.new_function)()
        self.assertEqual(
            self.messages,
            [
                'The function "test.new_function" is using its deprecated '
                'version and will expire in version "Beryllium".'
            ],
        )

    def test_with_deprecated_found(self):
        """
        Test with_deprecated should not raise an exception, if a same name
        function with the "_" prefix is implemented, but should use
        an old version instead, if "use_deprecated" is requested.

        :return:
        """
        self.globs["__opts__"]["use_deprecated"] = ["test.new_function"]
        self.globs["_new_function"] = self.old_function
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        self.assertEqual(depr(self.new_function)(), self.old_function())
        log_msg = [
            'The function "test.new_function" is using its deprecated version '
            'and will expire in version "Beryllium".'
        ]
        self.assertEqual(self.messages, log_msg)

    def test_with_deprecated_found_in_pillar(self):
        """
        Test with_deprecated should not raise an exception, if a same name
        function with the "_" prefix is implemented, but should use
        an old version instead, if "use_deprecated" is requested.

        :return:
        """
        self.globs["__pillar__"]["use_deprecated"] = ["test.new_function"]
        self.globs["_new_function"] = self.old_function
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        self.assertEqual(depr(self.new_function)(), self.old_function())
        log_msg = [
            'The function "test.new_function" is using its deprecated version '
            'and will expire in version "Beryllium".'
        ]
        self.assertEqual(self.messages, log_msg)

    def test_with_deprecated_found_eol(self):
        """
        Test with_deprecated should raise an exception, if a same name
        function with the "_" prefix is implemented, "use_deprecated" is requested
        and EOL is reached.

        :return:
        """
        self.globs["__opts__"]["use_deprecated"] = ["test.new_function"]
        self.globs["_new_function"] = self.old_function
        depr = decorators.with_deprecated(self.globs, "Helium")
        depr._curr_version = self._mk_version("Beryllium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.new_function)()
        self.assertEqual(
            self.messages,
            [
                'Although function "new_function" is called, an alias "new_function" is'
                " configured as its deprecated version. The lifetime of the function"
                ' "new_function" expired. Please use its successor "new_function"'
                " instead."
            ],
        )

    def test_with_deprecated_found_eol_in_pillar(self):
        """
        Test with_deprecated should raise an exception, if a same name
        function with the "_" prefix is implemented, "use_deprecated" is requested
        and EOL is reached.

        :return:
        """
        self.globs["__pillar__"]["use_deprecated"] = ["test.new_function"]
        self.globs["_new_function"] = self.old_function
        depr = decorators.with_deprecated(self.globs, "Helium")
        depr._curr_version = self._mk_version("Beryllium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.new_function)()
        self.assertEqual(
            self.messages,
            [
                'Although function "new_function" is called, an alias "new_function" is'
                " configured as its deprecated version. The lifetime of the function"
                ' "new_function" expired. Please use its successor "new_function"'
                " instead."
            ],
        )

    def test_with_deprecated_no_conf(self):
        """
        Test with_deprecated should not raise an exception, if a same name
        function with the "_" prefix is implemented, but should use
        a new version instead, if "use_deprecated" is not requested.

        :return:
        """
        self.globs["_new_function"] = self.old_function
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        self.assertEqual(depr(self.new_function)(), self.new_function())
        self.assertFalse(self.messages)

    def test_with_deprecated_with_name(self):
        """
        Test with_deprecated should not raise an exception, if a different name
        function is implemented and specified with the "with_name" parameter,
        but should use an old version instead and log a warning log message.

        :return:
        """
        self.globs["__opts__"]["use_deprecated"] = ["test.new_function"]
        depr = decorators.with_deprecated(
            self.globs, "Beryllium", with_name="old_function"
        )
        depr._curr_version = self._mk_version("Helium")[1]
        self.assertEqual(depr(self.new_function)(), self.old_function())
        self.assertEqual(
            self.messages,
            [
                'The function "old_function" is deprecated and will expire in version'
                ' "Beryllium". Use its successor "new_function" instead.'
            ],
        )

    def test_with_deprecated_with_name_eol(self):
        """
        Test with_deprecated should raise an exception, if a different name
        function is implemented and specified with the "with_name" parameter
        and EOL is reached.

        :return:
        """
        self.globs["__opts__"]["use_deprecated"] = ["test.new_function"]
        depr = decorators.with_deprecated(
            self.globs, "Helium", with_name="old_function"
        )
        depr._curr_version = self._mk_version("Beryllium")[1]
        with self.assertRaises(CommandExecutionError):
            depr(self.new_function)()
        self.assertEqual(
            self.messages,
            [
                'Although function "new_function" is called, '
                'an alias "old_function" is configured as its deprecated version. '
                'The lifetime of the function "old_function" expired. '
                'Please use its successor "new_function" instead.'
            ],
        )

    def test_with_deprecated_opt_in_default(self):
        """
        Test with_deprecated using opt-in policy,
        where newer function is not used, unless configured.

        :return:
        """
        depr = decorators.with_deprecated(
            self.globs, "Beryllium", policy=decorators._DeprecationDecorator.OPT_IN
        )
        depr._curr_version = self._mk_version("Helium")[1]
        assert depr(self.new_function)() == self.old_function()
        assert self.messages == [
            'The function "test.new_function" is using its '
            'deprecated version and will expire in version "Beryllium".'
        ]

    def test_with_deprecated_opt_in_use_superseded(self):
        """
        Test with_deprecated using opt-in policy,
        where newer function is used as per configuration.

        :return:
        """
        self.globs["__opts__"]["use_superseded"] = ["test.new_function"]
        depr = decorators.with_deprecated(
            self.globs, "Beryllium", policy=decorators._DeprecationDecorator.OPT_IN
        )
        depr._curr_version = self._mk_version("Helium")[1]
        assert depr(self.new_function)() == self.new_function()
        assert not self.messages

    def test_with_deprecated_opt_in_use_superseded_in_pillar(self):
        """
        Test with_deprecated using opt-in policy,
        where newer function is used as per configuration.

        :return:
        """
        self.globs["__pillar__"]["use_superseded"] = ["test.new_function"]
        depr = decorators.with_deprecated(
            self.globs, "Beryllium", policy=decorators._DeprecationDecorator.OPT_IN
        )
        depr._curr_version = self._mk_version("Helium")[1]
        assert depr(self.new_function)() == self.new_function()
        assert not self.messages

    def test_with_deprecated_opt_in_use_superseded_and_deprecated(self):
        """
        Test with_deprecated misconfiguration.

        :return:
        """
        self.globs["__opts__"]["use_deprecated"] = ["test.new_function"]
        self.globs["__opts__"]["use_superseded"] = ["test.new_function"]
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        with self.assertRaises(SaltConfigurationError):
            assert depr(self.new_function)() == self.new_function()

    def test_with_deprecated_opt_in_use_superseded_and_deprecated_in_pillar(self):
        """
        Test with_deprecated misconfiguration.

        :return:
        """
        self.globs["__pillar__"]["use_deprecated"] = ["test.new_function"]
        self.globs["__pillar__"]["use_superseded"] = ["test.new_function"]
        depr = decorators.with_deprecated(self.globs, "Beryllium")
        depr._curr_version = self._mk_version("Helium")[1]
        with self.assertRaises(SaltConfigurationError):
            assert depr(self.new_function)() == self.new_function()

    def test_allow_one_of(self):
        """
        Test allow_one_of properly does not error when only one of the
        required arguments is passed.

        :return:
        """
        allow_one_of = decorators.allow_one_of("arg1", "arg2", "arg3")
        assert allow_one_of(self.arg_function)(arg1="good") == self.arg_function(
            arg1="good"
        )

    def test_allow_one_of_succeeds_when_no_arguments_supplied(self):
        """
        Test allow_one_of properly does not error when none of the allowed
        arguments are supplied.

        :return:
        """
        allow_one_of = decorators.allow_one_of("arg1", "arg2", "arg3")
        assert allow_one_of(self.arg_function)() == self.arg_function()

    def test_allow_one_of_raises_error_when_multiple_allowed_arguments_supplied(self):
        """
        Test allow_one_of properly does not error when only one of the
        required arguments is passed.

        :return:
        """
        allow_one_of = decorators.allow_one_of("arg1", "arg2", "arg3")
        with self.assertRaises(SaltInvocationError):
            allow_one_of(self.arg_function)(arg1="good", arg2="bad")

    def test_require_one_of(self):
        """
        Test require_one_of properly does not error when only one of the
        required arguments is passed.

        :return:
        """
        require_one_of = decorators.require_one_of("arg1", "arg2", "arg3")
        assert require_one_of(self.arg_function)(arg1="good") == self.arg_function(
            arg1="good"
        )

    def test_require_one_of_raises_error_when_none_of_allowed_arguments_supplied(self):
        """
        Test require_one_of properly raises an error when none of the required
        arguments are supplied.

        :return:
        """
        require_one_of = decorators.require_one_of("arg1", "arg2", "arg3")
        with self.assertRaises(SaltInvocationError):
            require_one_of(self.arg_function)()

    def test_require_one_of_raises_error_when_multiple_allowed_arguments_supplied(self):
        """
        Test require_one_of properly raises an error when multiples of the
        allowed arguments are supplied.

        :return:
        """
        require_one_of = decorators.require_one_of("arg1", "arg2", "arg3")
        with self.assertRaises(SaltInvocationError):
            require_one_of(self.new_function)(arg1="good", arg2="bad")

    def test_with_depreciated_should_wrap_function(self):
        wrapped = decorators.with_deprecated({}, "Beryllium")(self.old_function)
        assert wrapped.__module__ == self.old_function.__module__

    def test_is_deprecated_should_wrap_function(self):
        wrapped = decorators.is_deprecated({}, "Beryllium")(self.old_function)
        assert wrapped.__module__ == self.old_function.__module__

    def test_ensure_unicode_args_should_wrap_function(self):
        wrapped = decorators.ensure_unicode_args(self.old_function)
        assert wrapped.__module__ == self.old_function.__module__

    def test_ignores_kwargs_should_wrap_function(self):
        wrapped = decorators.ignores_kwargs("foo", "bar")(self.old_function)
        assert wrapped.__module__ == self.old_function.__module__

    def test_memoize_should_wrap_function(self):
        wrapped = decorators.memoize(self.old_function)
        assert wrapped.__module__ == self.old_function.__module__

    def timing_should_wrap_function(self):
        wrapped = decorators.timing(self.old_function)
        assert wrapped.__module__ == self.old_function.__module__


class DependsDecoratorTest(TestCase):
    def function(self):
        return "foo"

    def test_depends_get_previous_frame(self):
        """
        Confirms that we're not grabbing the entire stack every time the
        depends decorator is invoked.
        """
        # Simply using True as a conditon; we aren't testing the dependency,
        # but rather the functions called within the decorator.
        dep = decorators.depends(True)

        # By mocking both inspect.stack and inspect.currentframe with
        # MagicMocks that return themselves, we don't affect normal operation
        # of the decorator, and at the same time we get to peek at whether or
        # not either was called.
        stack_mock = MagicMock(return_value=inspect.stack)
        currentframe_mock = MagicMock(return_value=inspect.currentframe)

        with patch.object(inspect, "stack", stack_mock), patch.object(
            inspect, "currentframe", currentframe_mock
        ):
            dep(self.function)()

        stack_mock.assert_not_called()
        currentframe_mock.assert_called_once_with()



import logging

import pytest

import salt.utils.azurearm as azurearm
from tests.support.unit import TestCase

# Azure libs
# pylint: disable=import-error
HAS_LIBS = False
try:
    import azure.mgmt.compute.models  # pylint: disable=unused-import
    import azure.mgmt.network.models  # pylint: disable=unused-import

    HAS_LIBS = True
except ImportError:
    pass

# pylint: enable=import-error

log = logging.getLogger(__name__)

MOCK_CREDENTIALS = {
    "client_id": "CLIENT_ID",
    "secret": "SECRET",
    "subscription_id": "SUBSCRIPTION_ID",
    "tenant": "TENANT",
}


@pytest.mark.skipif(
    HAS_LIBS is False, reason="The azure.mgmt.network module must be installed."
)
class AzureRmUtilsTestCase(TestCase):
    def test_create_object_model_vnet(self):
        module_name = "network"
        object_name = "VirtualNetwork"
        vnet = {
            "address_space": {"address_prefixes": ["10.0.0.0/8"]},
            "enable_ddos_protection": False,
            "enable_vm_protection": True,
            "tags": {"contact_name": "Elmer Fudd Gantry"},
        }
        model = azurearm.create_object_model(module_name, object_name, **vnet)
        self.assertEqual(vnet, model.as_dict())

    def test_create_object_model_nic_ref(self):
        module_name = "compute"
        object_name = "NetworkInterfaceReference"
        ref = {
            "id": "/subscriptions/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/resourceGroups/rg/providers/Microsoft.Network/networkInterfaces/nic",
            "primary": False,
        }
        model = azurearm.create_object_model(module_name, object_name, **ref)
        self.assertEqual(ref, model.as_dict())



import os
import textwrap

import pytest

import salt.utils.platform
import salt.utils.stringutils

pytestmark = [
    pytest.mark.windows_whitelisted,
]


class BlockreplaceParts:
    marker_start = "# start"
    marker_end = "# end"
    content = textwrap.dedent(
        """\
        Line 1 of block
        Line 2 of block
        """
    )
    without_block = textwrap.dedent(
        """\
        Hello world!

        # comment here
        """
    )
    with_non_matching_block = textwrap.dedent(
        """\
        Hello world!

        # start
        No match here
        # end
        # comment here
        """
    )
    with_non_matching_block_and_marker_end_not_after_newline = textwrap.dedent(
        """\
        Hello world!

        # start
        No match here# end
        # comment here
        """
    )
    with_matching_block = textwrap.dedent(
        """\
        Hello world!

        # start
        Line 1 of block
        Line 2 of block
        # end
        # comment here
        """
    )
    with_matching_block_and_extra_newline = textwrap.dedent(
        """\
        Hello world!

        # start
        Line 1 of block
        Line 2 of block

        # end
        # comment here
        """
    )
    with_matching_block_and_marker_end_not_after_newline = textwrap.dedent(
        """\
        Hello world!

        # start
        Line 1 of block
        Line 2 of block# end
        # comment here
        """
    )
    content_explicit_posix_newlines = "Line 1 of block\nLine 2 of block\n"
    content_explicit_windows_newlines = "Line 1 of block\r\nLine 2 of block\r\n"
    without_block_explicit_posix_newlines = "Hello world!\n\n# comment here\n"
    without_block_explicit_windows_newlines = "Hello world!\r\n\r\n# comment here\r\n"
    with_block_prepended_explicit_posix_newlines = (
        "# start\n"
        "Line 1 of block\n"
        "Line 2 of block\n"
        "# end\n"
        "Hello world!\n\n"
        "# comment here\n"
    )
    with_block_prepended_explicit_windows_newlines = (
        "# start\r\n"
        "Line 1 of block\r\n"
        "Line 2 of block\r\n"
        "# end\r\n"
        "Hello world!\r\n\r\n"
        "# comment here\r\n"
    )
    with_block_appended_explicit_posix_newlines = (
        "Hello world!\n\n"
        "# comment here\n"
        "# start\n"
        "Line 1 of block\n"
        "Line 2 of block\n"
        "# end\n"
    )
    with_block_appended_explicit_windows_newlines = (
        "Hello world!\r\n\r\n"
        "# comment here\r\n"
        "# start\r\n"
        "Line 1 of block\r\n"
        "Line 2 of block\r\n"
        "# end\r\n"
    )


def strip_ending_linebreak_ids(value):
    return f"strip_ending_linebreak={value}"


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_prepend(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when prepend_if_not_found=True and block doesn't
    exist in file.
    """
    name = tmp_path / "testfile"
    expected = (
        BlockreplaceParts.marker_start
        + "\n"
        + BlockreplaceParts.content
        + BlockreplaceParts.marker_end
        + "\n"
        + BlockreplaceParts.without_block
    )
    name.write_text(BlockreplaceParts.without_block)

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
    )
    assert ret.result is True
    assert ret.changes
    contents = name.read_text()
    assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
    )
    assert ret.result is True
    assert not ret.changes
    contents = name.read_text()
    assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_prepend_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when prepend_if_not_found=True and block doesn't
    exist in file. Test with append_newline explicitly set to True.
    """
    name = tmp_path / "testfile"
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content
            + BlockreplaceParts.marker_end
            + "\n"
            + BlockreplaceParts.without_block
        )
    else:
        expected = (
            BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content
            + "\n"
            + BlockreplaceParts.marker_end
            + "\n"
            + BlockreplaceParts.without_block
        )
    name.write_text(BlockreplaceParts.without_block)

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
        append_newline=True,
    )
    assert ret.result is True
    assert ret.changes
    contents = name.read_text()
    assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
        append_newline=True,
    )
    assert ret.result is True
    assert not ret.changes
    contents = name.read_text()
    assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_prepend_no_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when prepend_if_not_found=True and block doesn't
    exist in file. Test with append_newline explicitly set to False.
    """
    name = tmp_path / "testfile"
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content.rstrip("\r\n")
            + BlockreplaceParts.marker_end
            + "\n"
            + BlockreplaceParts.without_block
        )
    else:
        expected = (
            BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content
            + BlockreplaceParts.marker_end
            + "\n"
            + BlockreplaceParts.without_block
        )
    name.write_text(BlockreplaceParts.without_block)

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
        append_newline=False,
    )
    assert ret.result is True
    assert ret.changes
    contents = name.read_text()
    assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
        append_newline=False,
    )
    assert ret.result is True
    assert not ret.changes
    contents = name.read_text()
    assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_append(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when append_if_not_found=True and block doesn't
    exist in file.
    """
    name = tmp_path / "testfile"
    expected = (
        BlockreplaceParts.without_block
        + BlockreplaceParts.marker_start
        + "\n"
        + BlockreplaceParts.content
        + BlockreplaceParts.marker_end
        + "\n"
    )
    name.write_text(BlockreplaceParts.without_block)

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
    )
    assert ret.result is True
    assert ret.changes
    contents = name.read_text()
    assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
    )
    assert ret.result is True
    assert not ret.changes
    contents = name.read_text()
    assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_append_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when append_if_not_found=True and block doesn't
    exist in file. Test with append_newline explicitly set to True.
    """
    name = tmp_path / "testfile"
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.without_block
            + BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content
            + BlockreplaceParts.marker_end
            + "\n"
        )
    else:
        expected = (
            BlockreplaceParts.without_block
            + BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content
            + "\n"
            + BlockreplaceParts.marker_end
            + "\n"
        )
    name.write_text(BlockreplaceParts.without_block)

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
        append_newline=True,
    )
    assert ret.result is True
    assert ret.changes
    contents = name.read_text()
    assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
        append_newline=True,
    )
    assert ret.result is True
    assert not ret.changes
    contents = name.read_text()
    assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_append_no_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when append_if_not_found=True and block doesn't
    exist in file. Test with append_newline explicitly set to False.
    """
    name = tmp_path / "testfile"
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.without_block
            + BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content.rstrip("\r\n")
            + BlockreplaceParts.marker_end
            + "\n"
        )
    else:
        expected = (
            BlockreplaceParts.without_block
            + BlockreplaceParts.marker_start
            + "\n"
            + BlockreplaceParts.content
            + BlockreplaceParts.marker_end
            + "\n"
        )
    name.write_text(BlockreplaceParts.without_block)

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
        append_newline=False,
    )
    assert ret.result is True
    assert ret.changes
    contents = name.read_text()
    assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
        append_newline=False,
    )
    assert ret.result is True
    assert not ret.changes
    contents = name.read_text()
    assert contents == expected


def line_breaks_ids(value):
    return f"line_breaks={value}"


@pytest.mark.parametrize("line_breaks", ("windows", "posix"), ids=line_breaks_ids)
def test_prepend_auto_line_separator(file, tmp_path, line_breaks):
    """
    This tests the line separator auto-detection when prepending the block
    """
    name = tmp_path / "testfile"
    if line_breaks == "posix":
        # POSIX newlines to Windows newlines
        name.write_text(BlockreplaceParts.without_block_explicit_windows_newlines)
        content = BlockreplaceParts.content_explicit_posix_newlines
        expected = BlockreplaceParts.with_block_prepended_explicit_windows_newlines
    else:
        # Windows newlines to POSIX newlines
        name.write_text(BlockreplaceParts.without_block_explicit_posix_newlines)
        content = BlockreplaceParts.content_explicit_windows_newlines
        expected = BlockreplaceParts.with_block_prepended_explicit_posix_newlines

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        prepend_if_not_found=True,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize("line_breaks", ("windows", "posix"), ids=line_breaks_ids)
def test_append_auto_line_separator(file, tmp_path, line_breaks):
    """
    This tests the line separator auto-detection when appending the block
    """
    name = tmp_path / "testfile"
    if line_breaks == "posix":
        # POSIX newlines to Windows newlines
        name.write_text(BlockreplaceParts.without_block_explicit_windows_newlines)
        content = BlockreplaceParts.content_explicit_posix_newlines
        expected = BlockreplaceParts.with_block_appended_explicit_windows_newlines
    else:
        # Windows newlines to POSIX newlines
        name.write_text(BlockreplaceParts.without_block_explicit_posix_newlines)
        content = BlockreplaceParts.content_explicit_windows_newlines
        expected = BlockreplaceParts.with_block_appended_explicit_posix_newlines

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_if_not_found=True,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_non_matching_block(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when block exists but its contents are not a match.
    """
    name = tmp_path / "testfile"
    name.write_text(BlockreplaceParts.with_non_matching_block)
    expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_non_matching_block_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when block exists but its contents are not a
    match. Test with append_newline explicitly set to True.
    """
    name = tmp_path / "testfile"
    name.write_text(BlockreplaceParts.with_non_matching_block)
    if strip_ending_linebreak:
        expected = BlockreplaceParts.with_matching_block
    else:
        expected = BlockreplaceParts.with_matching_block_and_extra_newline

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_non_matching_block_no_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when block exists but its contents are not a
    match. Test with append_newline explicitly set to False.
    """
    name = tmp_path / "testfile"
    name.write_text(BlockreplaceParts.with_non_matching_block)
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
        )
    else:
        expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_non_matching_block_and_marker_not_after_newline(
    file, tmp_path, strip_ending_linebreak
):
    """
    Test blockreplace when block exists but its contents are not a
    match, and the marker_end is not directly preceded by a newline.
    """
    name = tmp_path / "testfile"
    name.write_text(
        BlockreplaceParts.with_non_matching_block_and_marker_end_not_after_newline
    )
    expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_non_matching_block_and_marker_not_after_newline_append_newline(
    file, tmp_path, strip_ending_linebreak
):
    """
    Test blockreplace when block exists but its contents are not a match,
    and the marker_end is not directly preceded by a newline. Test with
    append_newline explicitly set to True.
    """
    name = tmp_path / "testfile"
    name.write_text(
        BlockreplaceParts.with_non_matching_block_and_marker_end_not_after_newline
    )
    if strip_ending_linebreak:
        expected = BlockreplaceParts.with_matching_block
    else:
        expected = BlockreplaceParts.with_matching_block_and_extra_newline

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_non_matching_block_and_marker_not_after_newline_no_append_newline(
    file, tmp_path, strip_ending_linebreak
):
    """
    Test blockreplace when block exists but its contents are not a match,
    and the marker_end is not directly preceded by a newline. Test with
    append_newline explicitly set to False.
    """
    name = tmp_path / "testfile"
    name.write_text(
        BlockreplaceParts.with_non_matching_block_and_marker_end_not_after_newline
    )
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
        )
    else:
        expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_matching_block(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when block exists and its contents are a match. No
    changes should be made.
    """
    name = tmp_path / "testfile"
    name.write_text(BlockreplaceParts.with_matching_block)
    expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_matching_block_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when block exists and its contents are a match. Test
    with append_newline explicitly set to True. This will result in an
    extra newline when the content ends in a newline, and will not when the
    content does not end in a newline.
    """
    name = tmp_path / "testfile"
    name.write_text(BlockreplaceParts.with_matching_block)
    if strip_ending_linebreak:
        expected = BlockreplaceParts.with_matching_block
    else:
        expected = BlockreplaceParts.with_matching_block_and_extra_newline

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    if strip_ending_linebreak:
        assert not ret.changes
    else:
        assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_matching_block_no_append_newline(file, tmp_path, strip_ending_linebreak):
    """
    Test blockreplace when block exists and its contents are a match. Test
    with append_newline explicitly set to False. This will result in the
    marker_end not being directly preceded by a newline when the content
    does not end in a newline.
    """
    name = tmp_path / "testfile"
    name.write_text(BlockreplaceParts.with_matching_block)
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
        )
    else:
        expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    if strip_ending_linebreak:
        assert ret.changes
    else:
        assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_matching_block_and_marker_not_after_newline(
    file, tmp_path, strip_ending_linebreak
):
    """
    Test blockreplace when block exists and its contents are a match, but
    the marker_end is not directly preceded by a newline.
    """
    name = tmp_path / "testfile"
    name.write_text(
        BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
    )
    expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_matching_block_and_marker_not_after_newline_append_newline(
    file, tmp_path, strip_ending_linebreak
):
    """
    Test blockreplace when block exists and its contents are a match, but
    the marker_end is not directly preceded by a newline. Test with
    append_newline explicitly set to True. This will result in an extra
    newline when the content ends in a newline, and will not when the
    content does not end in a newline.
    """
    name = tmp_path / "testfile"
    name.write_text(
        BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
    )
    if strip_ending_linebreak:
        expected = BlockreplaceParts.with_matching_block
    else:
        expected = BlockreplaceParts.with_matching_block_and_extra_newline

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=True,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


@pytest.mark.parametrize(
    "strip_ending_linebreak", (False, True), ids=strip_ending_linebreak_ids
)
def test_matching_block_and_marker_not_after_newline_no_append_newline(
    file, tmp_path, strip_ending_linebreak
):
    """
    Test blockreplace when block exists and its contents are a match, but
    the marker_end is not directly preceded by a newline. Test with
    append_newline explicitly set to False.
    """
    name = tmp_path / "testfile"
    name.write_text(
        BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
    )
    if strip_ending_linebreak:
        expected = (
            BlockreplaceParts.with_matching_block_and_marker_end_not_after_newline
        )
    else:
        expected = BlockreplaceParts.with_matching_block

    # content ends in newline
    content = BlockreplaceParts.content
    if strip_ending_linebreak:
        # content does not end in newline
        content = content.rstrip("\r\n")

    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    if strip_ending_linebreak:
        assert not ret.changes
    else:
        assert ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected

    # Re-run state, no changes should be made
    ret = file.blockreplace(
        name=str(name),
        content=content,
        marker_start=BlockreplaceParts.marker_start,
        marker_end=BlockreplaceParts.marker_end,
        append_newline=False,
    )
    assert ret.result is True
    assert not ret.changes
    # The following assertion fails on windows under functional tests.
    # Asserting against, ret.change is enough
    if not salt.utils.platform.is_windows():
        # Read bytes to preserve line endings since in text mode, '\r\n' is converted to '\n'
        contents = salt.utils.stringutils.to_unicode(name.read_bytes())
        assert contents == expected


def test_issue_49043(file, tmp_path, state_tree):
    """
    Test blockreplace with jinja template and unicode strings in the context
    """
    if salt.utils.platform.is_windows() and os.environ.get("PYTHONUTF8", "0") == "0":
        pytest.skip("Test will fail if PYTHONUTF8=1 is not set on windows")
    name = tmp_path / "testfile"
    name.touch()
    unicode_string = "äöü"
    expected = textwrap.dedent(
        """\
        #-- start managed zone --
        {}
        #-- end managed zone --
        """.format(
            unicode_string
        )
    )
    with pytest.helpers.temp_file(
        "issue-49043", directory=state_tree, contents="{{ unicode_string }}"
    ):
        ret = file.blockreplace(
            name=str(name),
            source="salt://issue-49043",
            append_if_not_found=True,
            template="jinja",
            context={"unicode_string": unicode_string},
        )
        assert ret.result is True
        assert name.read_text() == expected



import pytest

import salt.utils.platform

if salt.utils.platform.is_windows():
    from salt.platform import win

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.mark.parametrize(
    "command, expected",
    [
        ("whoami", "cmd.exe /c whoami"),
        ("cmd /c hostname", "cmd.exe /c cmd /c hostname"),
        ("echo foo", "cmd.exe /c echo foo"),
        ('cmd /c "echo foo"', 'cmd.exe /c cmd /c "echo foo"'),
        ("icacls 'C:\\Program Files'", "cmd.exe /c icacls 'C:\\Program Files'"),
        (
            "icacls 'C:\\Program Files' && echo 1",
            "cmd.exe /c icacls 'C:\\Program Files' && echo 1",
        ),
        (
            ["secedit", "/export", "/cfg", "C:\\A Path\\with\\a\\space"],
            'cmd.exe /c secedit /export /cfg "C:\\A Path\\with\\a\\space"',
        ),
        (
            ["C:\\a space\\a space.bat", "foo foo", "bar bar"],
            'cmd.exe /c "C:\\a space\\a space.bat" "foo foo" "bar bar"',
        ),
        (
            '''echo "&<>[]|{}^=;!'+,`~ "''',
            '''cmd.exe /c echo "&<>[]|{}^=;!'+,`~ "''',
        ),
    ],
)
def test_prepend_cmd(command, expected):
    """
    Test that the command is prepended with "cmd /c" and quoted
    """
    win_shell = "cmd.exe"
    result = win.prepend_cmd(win_shell, command)
    assert result == expected



"""
Run remote execution commands via the local client
"""

import salt.client


def cmd(name, tgt, func, arg=(), tgt_type="glob", ret="", kwarg=None, **kwargs):
    """
    Execute a remote execution command

    USAGE:

    .. code-block:: yaml

        run_remote_ex:
          local.cmd:
            - tgt: '*'
            - func: test.ping

        run_remote_ex:
          local.cmd:
            - tgt: '*'
            - func: test.sleep
            - arg:
              - 30

        run_remote_ex:
          local.cmd:
            - tgt: '*'
            - func: test.sleep
            - kwarg:
              length: 30
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    with salt.client.get_local_client(mopts=__opts__) as client:
        jid = client.cmd_async(
            tgt, func, arg, tgt_type=tgt_type, ret=ret, kwarg=kwarg, **kwargs
        )
        ret["changes"]["jid"] = jid
    return ret



import pytest

import salt.utils.versions

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]

# The swarm module need the docker-py library installed
pytest.importorskip("docker")


@pytest.fixture
def docker_version(shell, grains):
    ret = shell.run("docker", "--version")
    assert ret.returncode == 0
    return salt.utils.versions.Version(ret.stdout.split(",")[0].split()[-1].strip())


@pytest.fixture
def swarm(modules):
    try:
        yield modules.swarm
    finally:
        modules.swarm.leave_swarm(force=True)


def test_swarm_init(swarm, docker_version, grains):
    """
    check that swarm.swarm_init works when a swarm exists
    """
    ret = swarm.swarm_init("127.0.0.1", "0.0.0.0", False)
    if "Error" in ret:
        if (
            ret["Error"]
            == "--live-restore daemon configuration is incompatible with swarm mode"
        ):
            if grains["os"] == "Fedora" and int(grains["osmajorrelease"]) >= 37:
                pytest.skip(
                    "The 'swarm' module needs to be updated to support the "
                    f"system docker {docker_version} package"
                )
    assert "Error" not in ret
    for key in ("Comment", "Tokens"):
        assert key in ret
    assert ret["Comment"].startswith("Docker swarm has been initialized on ")

    ret = swarm.swarm_init("127.0.0.1", "0.0.0.0", False)
    expected = {
        "Comment": (
            'This node is already part of a swarm. Use "docker swarm leave" to'
            " leave this swarm and join another one."
        ),
        "result": False,
    }
    assert ret == expected



"""
Send a message card to Microsoft Teams
======================================

This state is useful for sending messages to Teams during state runs.

.. versionadded:: 2017.7.0

.. code-block:: yaml

    teams-message:
      msteams.post_card:
        - message: 'This state was executed successfully.'
        - hook_url:  https://outlook.office.com/webhook/837

The hook_url can be specified in the master or minion configuration like below:

.. code-block:: yaml

    msteams:
      hook_url: https://outlook.office.com/webhook/837
"""

from salt.exceptions import SaltInvocationError


def __virtual__():
    """
    Only load if the msteams module is available in __salt__
    """
    if "msteams.post_card" in __salt__:
        return "msteams"
    return (False, "msteams module could not be loaded")


def post_card(name, message, hook_url=None, title=None, theme_color=None):
    """
    Send a message to a Microsft Teams channel

    .. code-block:: yaml

        send-msteams-message:
          msteams.post_card:
            - message: 'This state was executed successfully.'
            - hook_url: https://outlook.office.com/webhook/837

    The following parameters are required:

    message
        The message that is to be sent to the MS Teams channel.

    The following parameters are optional:

    hook_url
        The webhook URL given configured in Teams interface,
        if not specified in the configuration options of master or minion.
    title
        The title for the card posted to the channel
    theme_color
        A hex code for the desired highlight color
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if __opts__["test"]:
        ret["comment"] = "The following message is to be sent to Teams: {}".format(
            message
        )
        ret["result"] = None
        return ret

    if not message:
        ret["comment"] = f"Teams message is missing: {message}"
        return ret

    try:
        result = __salt__["msteams.post_card"](
            message=message,
            hook_url=hook_url,
            title=title,
            theme_color=theme_color,
        )
    except SaltInvocationError as sie:
        ret["comment"] = f"Failed to send message ({sie}): {name}"
    else:
        if isinstance(result, bool) and result:
            ret["result"] = True
            ret["comment"] = f"Sent message: {name}"
        else:
            ret["comment"] = "Failed to send message ({}): {}".format(
                result["message"], name
            )

    return ret



"""
DigitalOcean Cloud Module
=========================

The DigitalOcean cloud module is used to control access to the DigitalOcean VPS system.

Use of this module requires a requires a ``personal_access_token``, an ``ssh_key_file``,
and at least one SSH key name in ``ssh_key_names``. More ``ssh_key_names`` can be added
by separating each key with a comma. The ``personal_access_token`` can be found in the
DigitalOcean web interface in the "Apps & API" section. The SSH key name can be found
under the "SSH Keys" section.

.. code-block:: yaml

    # Note: This example is for /etc/salt/cloud.providers or any file in the
    # /etc/salt/cloud.providers.d/ directory.

    my-digital-ocean-config:
      personal_access_token: xxx
      ssh_key_file: /path/to/ssh/key/file
      ssh_key_names: my-key-name,my-key-name-2
      driver: digitalocean

:depends: requests
"""

import decimal
import logging
import os
import pprint
import time

import salt.config as config
import salt.utils.cloud
import salt.utils.files
import salt.utils.json
import salt.utils.stringutils
from salt.exceptions import (
    SaltCloudConfigError,
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
    SaltInvocationError,
)

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "digitalocean"
__virtual_aliases__ = ("digital_ocean", "do")


# Only load in this module if the DIGITALOCEAN configurations are in place
def __virtual__():
    """
    Check for DigitalOcean configurations
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        opts=__opts__,
        provider=_get_active_provider_name() or __virtualname__,
        aliases=__virtual_aliases__,
        required_keys=("personal_access_token",),
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(__virtualname__, {"requests": HAS_REQUESTS})


def avail_locations(call=None):
    """
    Return a dict of all available VM locations on the cloud provider with
    relevant data
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    items = query(method="regions")
    ret = {}
    for region in items["regions"]:
        ret[region["name"]] = {}
        for item in region.keys():
            ret[region["name"]][item] = str(region[item])

    return ret


def avail_images(call=None):
    """
    Return a list of the images that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    fetch = True
    page = 1
    ret = {}

    while fetch:
        items = query(method="images", command="?page=" + str(page) + "&per_page=200")

        for image in items["images"]:
            ret[image["name"]] = {}
            for item in image.keys():
                ret[image["name"]][item] = image[item]

        page += 1
        try:
            fetch = "next" in items["links"]["pages"]
        except KeyError:
            fetch = False

    return ret


def avail_sizes(call=None):
    """
    Return a list of the image sizes that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    items = query(method="sizes", command="?per_page=100")
    ret = {}
    for size in items["sizes"]:
        ret[size["slug"]] = {}
        for item in size.keys():
            ret[size["slug"]][item] = str(size[item])

    return ret


def list_nodes(call=None):
    """
    Return a list of the VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )
    return _list_nodes()


def list_nodes_full(call=None, for_output=True):
    """
    Return a list of the VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )
    return _list_nodes(full=True, for_output=for_output)


def list_nodes_select(call=None):
    """
    Return a list of the VMs that are on the provider, with select fields
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full("function"),
        __opts__["query.selection"],
        call,
    )


def get_image(vm_):
    """
    Return the image object to use
    """
    images = avail_images()
    vm_image = config.get_cloud_config_value(
        "image", vm_, __opts__, search_global=False
    )
    if not isinstance(vm_image, str):
        vm_image = str(vm_image)

    for image in images:
        if vm_image in (
            images[image]["name"],
            images[image]["slug"],
            images[image]["id"],
        ):
            if images[image]["slug"] is not None:
                return images[image]["slug"]
            return int(images[image]["id"])
    raise SaltCloudNotFound(f"The specified image, '{vm_image}', could not be found.")


def get_size(vm_):
    """
    Return the VM's size. Used by create_node().
    """
    sizes = avail_sizes()
    vm_size = str(
        config.get_cloud_config_value("size", vm_, __opts__, search_global=False)
    )
    for size in sizes:
        if vm_size.lower() == sizes[size]["slug"]:
            return sizes[size]["slug"]
    raise SaltCloudNotFound(f"The specified size, '{vm_size}', could not be found.")


def get_location(vm_):
    """
    Return the VM's location
    """
    locations = avail_locations()
    vm_location = str(
        config.get_cloud_config_value("location", vm_, __opts__, search_global=False)
    )

    for location in locations:
        if vm_location in (locations[location]["name"], locations[location]["slug"]):
            return locations[location]["slug"]
    raise SaltCloudNotFound(
        f"The specified location, '{vm_location}', could not be found."
    )


def create_node(args):
    """
    Create a node
    """
    node = query(method="droplets", args=args, http_method="post")
    return node


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "digitalocean",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.info("Creating Cloud VM %s", vm_["name"])

    kwargs = {
        "name": vm_["name"],
        "size": get_size(vm_),
        "image": get_image(vm_),
        "region": get_location(vm_),
        "ssh_keys": [],
        "tags": [],
    }

    # backwards compat
    ssh_key_name = config.get_cloud_config_value(
        "ssh_key_name", vm_, __opts__, search_global=False
    )

    if ssh_key_name:
        kwargs["ssh_keys"].append(get_keyid(ssh_key_name))

    ssh_key_names = config.get_cloud_config_value(
        "ssh_key_names", vm_, __opts__, search_global=False, default=False
    )

    if ssh_key_names:
        for key in ssh_key_names.split(","):
            kwargs["ssh_keys"].append(get_keyid(key))

    key_filename = config.get_cloud_config_value(
        "ssh_key_file", vm_, __opts__, search_global=False, default=None
    )

    if key_filename is not None and not os.path.isfile(key_filename):
        raise SaltCloudConfigError(
            f"The defined key_filename '{key_filename}' does not exist"
        )

    if not __opts__.get("ssh_agent", False) and key_filename is None:
        raise SaltCloudConfigError(
            "The DigitalOcean driver requires an ssh_key_file and an ssh_key_name "
            "because it does not supply a root password upon building the server."
        )

    ssh_interface = config.get_cloud_config_value(
        "ssh_interface", vm_, __opts__, search_global=False, default="public"
    )

    if ssh_interface in ["private", "public"]:
        log.info("ssh_interface: Setting interface for ssh to %s", ssh_interface)
        kwargs["ssh_interface"] = ssh_interface
    else:
        raise SaltCloudConfigError(
            "The DigitalOcean driver requires ssh_interface to be defined as 'public'"
            " or 'private'."
        )

    private_networking = config.get_cloud_config_value(
        "private_networking",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if private_networking is not None:
        if not isinstance(private_networking, bool):
            raise SaltCloudConfigError(
                "'private_networking' should be a boolean value."
            )
        kwargs["private_networking"] = private_networking

    if not private_networking and ssh_interface == "private":
        raise SaltCloudConfigError(
            "The DigitalOcean driver requires ssh_interface if defined as 'private' "
            "then private_networking should be set as 'True'."
        )

    backups_enabled = config.get_cloud_config_value(
        "backups_enabled",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if backups_enabled is not None:
        if not isinstance(backups_enabled, bool):
            raise SaltCloudConfigError("'backups_enabled' should be a boolean value.")
        kwargs["backups"] = backups_enabled

    ipv6 = config.get_cloud_config_value(
        "ipv6",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if ipv6 is not None:
        if not isinstance(ipv6, bool):
            raise SaltCloudConfigError("'ipv6' should be a boolean value.")
        kwargs["ipv6"] = ipv6

    monitoring = config.get_cloud_config_value(
        "monitoring",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if monitoring is not None:
        if not isinstance(monitoring, bool):
            raise SaltCloudConfigError("'monitoring' should be a boolean value.")
        kwargs["monitoring"] = monitoring

    kwargs["tags"] = config.get_cloud_config_value(
        "tags", vm_, __opts__, search_global=False, default=False
    )

    userdata_file = config.get_cloud_config_value(
        "userdata_file", vm_, __opts__, search_global=False, default=None
    )
    if userdata_file is not None:
        try:
            with salt.utils.files.fopen(userdata_file, "r") as fp_:
                kwargs["user_data"] = salt.utils.cloud.userdata_template(
                    __opts__, vm_, salt.utils.stringutils.to_unicode(fp_.read())
                )
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Failed to read userdata from %s: %s", userdata_file, exc)

    create_dns_record = config.get_cloud_config_value(
        "create_dns_record",
        vm_,
        __opts__,
        search_global=False,
        default=None,
    )

    if create_dns_record:
        log.info("create_dns_record: will attempt to write DNS records")
        default_dns_domain = None
        dns_domain_name = vm_["name"].split(".")
        if len(dns_domain_name) > 2:
            log.debug(
                "create_dns_record: inferring default dns_hostname, dns_domain from"
                " minion name as FQDN"
            )
            default_dns_hostname = ".".join(dns_domain_name[:-2])
            default_dns_domain = ".".join(dns_domain_name[-2:])
        else:
            log.debug("create_dns_record: can't infer dns_domain from %s", vm_["name"])
            default_dns_hostname = dns_domain_name[0]

        dns_hostname = config.get_cloud_config_value(
            "dns_hostname",
            vm_,
            __opts__,
            search_global=False,
            default=default_dns_hostname,
        )
        dns_domain = config.get_cloud_config_value(
            "dns_domain",
            vm_,
            __opts__,
            search_global=False,
            default=default_dns_domain,
        )
        if dns_hostname and dns_domain:
            log.info(
                'create_dns_record: using dns_hostname="%s", dns_domain="%s"',
                dns_hostname,
                dns_domain,
            )

            def __add_dns_addr__(t, d):
                return post_dns_record(
                    dns_domain=dns_domain,
                    name=dns_hostname,
                    record_type=t,
                    record_data=d,
                )

            log.debug("create_dns_record: %s", __add_dns_addr__)
        else:
            log.error(
                "create_dns_record: could not determine dns_hostname and/or dns_domain"
            )
            raise SaltCloudConfigError(
                "'create_dns_record' must be a dict specifying \"domain\" "
                'and "hostname" or the minion name must be an FQDN.'
            )

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"]("requesting", kwargs, list(kwargs)),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    try:
        ret = create_node(kwargs)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error creating %s on DIGITALOCEAN\n\n"
            "The following exception was thrown when trying to "
            "run the initial deployment: %s",
            vm_["name"],
            exc,
            # Show the traceback if the debug logging level is enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False

    def __query_node_data(vm_name):
        data = show_instance(vm_name, "action")
        if not data:
            # Trigger an error in the wait_for_ip function
            return False
        if data["networks"].get("v4"):
            for network in data["networks"]["v4"]:
                if network["type"] == "public":
                    return data
        return False

    try:
        data = salt.utils.cloud.wait_for_ip(
            __query_node_data,
            update_args=(vm_["name"],),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc))

    if not vm_.get("ssh_host"):
        vm_["ssh_host"] = None

    # add DNS records, set ssh_host, default to first found IP, preferring IPv4 for ssh bootstrap script target
    addr_families, dns_arec_types = (("v4", "v6"), ("A", "AAAA"))
    arec_map = dict(list(zip(addr_families, dns_arec_types)))
    for facing, addr_family, ip_address in [
        (net["type"], family, net["ip_address"])
        for family in addr_families
        for net in data["networks"][family]
    ]:
        log.info('found %s IP%s interface for "%s"', facing, addr_family, ip_address)
        dns_rec_type = arec_map[addr_family]
        if facing == "public":
            if create_dns_record:
                __add_dns_addr__(dns_rec_type, ip_address)
        if facing == ssh_interface:
            if not vm_["ssh_host"]:
                vm_["ssh_host"] = ip_address

    if vm_["ssh_host"] is None:
        raise SaltCloudSystemExit(
            "No suitable IP addresses found for ssh minion bootstrapping: {}".format(
                repr(data["networks"])
            )
        )

    log.debug(
        "Found public IP address to use for ssh minion bootstrapping: %s",
        vm_["ssh_host"],
    )

    vm_["key_filename"] = key_filename
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)
    ret.update(data)

    log.info("Created Cloud VM '%s'", vm_["name"])
    log.debug("'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def query(
    method="droplets", droplet_id=None, command=None, args=None, http_method="get"
):
    """
    Make a web call to DigitalOcean
    """
    base_path = str(
        config.get_cloud_config_value(
            "api_root",
            get_configured_provider(),
            __opts__,
            search_global=False,
            default="https://api.digitalocean.com/v2",
        )
    )

    path = f"{base_path}/{method}/"

    if droplet_id:
        path += f"{droplet_id}/"

    if command:
        path += command

    if not isinstance(args, dict):
        args = {}

    personal_access_token = config.get_cloud_config_value(
        "personal_access_token",
        get_configured_provider(),
        __opts__,
        search_global=False,
    )

    data = salt.utils.json.dumps(args)

    requester = getattr(requests, http_method)
    request = requester(
        path,
        data=data,
        headers={
            "Authorization": "Bearer " + personal_access_token,
            "Content-Type": "application/json",
        },
        timeout=120,
    )
    if request.status_code > 299:
        raise SaltCloudSystemExit(
            "An error occurred while querying DigitalOcean. HTTP Code: {}  "
            "Error: '{}'".format(
                request.status_code,
                # request.read()
                request.text,
            )
        )

    log.debug(request.url)

    # success without data
    if request.status_code == 204:
        return True

    content = request.text

    result = salt.utils.json.loads(content)
    if result.get("status", "").lower() == "error":
        raise SaltCloudSystemExit(pprint.pformat(result.get("error_message", {})))

    return result


def script(vm_):
    """
    Return the script deployment object
    """
    deploy_script = salt.utils.cloud.os_script(
        config.get_cloud_config_value("script", vm_, __opts__),
        vm_,
        __opts__,
        salt.utils.cloud.salt_config_to_yaml(
            salt.utils.cloud.minion_config(__opts__, vm_)
        ),
    )
    return deploy_script


def show_instance(name, call=None):
    """
    Show the details from DigitalOcean concerning a droplet
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )
    node = _get_node(name)
    __utils__["cloud.cache_node"](node, _get_active_provider_name(), __opts__)
    return node


def _get_node(name):
    attempts = 10
    while attempts >= 0:
        try:
            return list_nodes_full(for_output=False)[name]
        except KeyError:
            attempts -= 1
            log.debug(
                "Failed to get the data for node '%s'. Remaining attempts: %s",
                name,
                attempts,
            )
            # Just a little delay between attempts...
            time.sleep(0.5)
    return {}


def list_keypairs(call=None):
    """
    Return a dict of all available VM locations on the cloud provider with
    relevant data
    """
    if call != "function":
        log.error("The list_keypairs function must be called with -f or --function.")
        return False

    fetch = True
    page = 1
    ret = {}

    while fetch:
        items = query(
            method="account/keys",
            command="?page=" + str(page) + "&per_page=100",
        )

        for key_pair in items["ssh_keys"]:
            name = key_pair["name"]
            if name in ret:
                raise SaltCloudSystemExit(
                    "A duplicate key pair name, '{}', was found in DigitalOcean's "
                    "key pair list. Please change the key name stored by DigitalOcean. "
                    "Be sure to adjust the value of 'ssh_key_file' in your cloud "
                    "profile or provider configuration, if necessary.".format(name)
                )
            ret[name] = {}
            for item in key_pair.keys():
                ret[name][item] = str(key_pair[item])

        page += 1
        try:
            fetch = "next" in items["links"]["pages"]
        except KeyError:
            fetch = False

    return ret


def show_keypair(kwargs=None, call=None):
    """
    Show the details of an SSH keypair
    """
    if call != "function":
        log.error("The show_keypair function must be called with -f or --function.")
        return False

    if not kwargs:
        kwargs = {}

    if "keyname" not in kwargs:
        log.error("A keyname is required.")
        return False

    keypairs = list_keypairs(call="function")
    keyid = keypairs[kwargs["keyname"]]["id"]
    log.debug("Key ID is %s", keyid)

    details = query(method="account/keys", command=keyid)

    return details


def import_keypair(kwargs=None, call=None):
    """
    Upload public key to cloud provider.
    Similar to EC2 import_keypair.

    .. versionadded:: 2016.11.0

    kwargs
        file(mandatory): public key file-name
        keyname(mandatory): public key name in the provider
    """
    with salt.utils.files.fopen(kwargs["file"], "r") as public_key_filename:
        public_key_content = salt.utils.stringutils.to_unicode(
            public_key_filename.read()
        )

    digitalocean_kwargs = {"name": kwargs["keyname"], "public_key": public_key_content}

    created_result = create_key(digitalocean_kwargs, call=call)
    return created_result


def create_key(kwargs=None, call=None):
    """
    Upload a public key
    """
    if call != "function":
        log.error("The create_key function must be called with -f or --function.")
        return False

    try:
        result = query(
            method="account",
            command="keys",
            args={"name": kwargs["name"], "public_key": kwargs["public_key"]},
            http_method="post",
        )
    except KeyError:
        log.info("`name` and `public_key` arguments must be specified")
        return False

    return result


def remove_key(kwargs=None, call=None):
    """
    Delete public key
    """
    if call != "function":
        log.error("The create_key function must be called with -f or --function.")
        return False

    try:
        result = query(
            method="account", command="keys/" + kwargs["id"], http_method="delete"
        )
    except KeyError:
        log.info("`id` argument must be specified")
        return False

    return result


def get_keyid(keyname):
    """
    Return the ID of the keyname
    """
    if not keyname:
        return None
    keypairs = list_keypairs(call="function")
    keyid = keypairs[keyname]["id"]
    if keyid:
        return keyid
    raise SaltCloudNotFound("The specified ssh key could not be found.")


def destroy(name, call=None):
    """
    Destroy a node. Will check termination protection and warn if enabled.

    CLI Example:

    .. code-block:: bash

        salt-cloud --destroy mymachine
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    data = show_instance(name, call="action")
    node = query(method="droplets", droplet_id=data["id"], http_method="delete")

    ## This is all terribly optomistic:
    # vm_ = get_vm_config(name=name)
    # delete_dns_record = config.get_cloud_config_value(
    #     'delete_dns_record', vm_, __opts__, search_global=False, default=None,
    # )
    # TODO: when _vm config data can be made available, we should honor the configuration settings,
    # but until then, we should assume stale DNS records are bad, and default behavior should be to
    # delete them if we can. When this is resolved, also resolve the comments a couple of lines below.
    delete_dns_record = True

    if not isinstance(delete_dns_record, bool):
        raise SaltCloudConfigError("'delete_dns_record' should be a boolean value.")
    # When the "to do" a few lines up is resolved, remove these lines and use the if/else logic below.
    log.debug("Deleting DNS records for %s.", name)
    destroy_dns_records(name)

    # Until the "to do" from line 754 is taken care of, we don't need this logic.
    # if delete_dns_record:
    #    log.debug('Deleting DNS records for %s.', name)
    #    destroy_dns_records(name)
    # else:
    #    log.debug('delete_dns_record : %s', delete_dns_record)
    #    for line in pprint.pformat(dir()).splitlines():
    #       log.debug('delete context: %s', line)

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    if __opts__.get("update_cachedir", False) is True:
        __utils__["cloud.delete_minion_cachedir"](
            name, _get_active_provider_name().split(":")[0], __opts__
        )

    return node


def post_dns_record(**kwargs):
    """
    Creates a DNS record for the given name if the domain is managed with DO.
    """
    if "kwargs" in kwargs:  # flatten kwargs if called via salt-cloud -f
        f_kwargs = kwargs["kwargs"]
        del kwargs["kwargs"]
        kwargs.update(f_kwargs)
    mandatory_kwargs = ("dns_domain", "name", "record_type", "record_data")
    for i in mandatory_kwargs:
        if kwargs[i]:
            pass
        else:
            error = '{}="{}" ## all mandatory args must be provided: {}'.format(
                i, kwargs[i], mandatory_kwargs
            )
            raise SaltInvocationError(error)

    domain = query(method="domains", droplet_id=kwargs["dns_domain"])

    if domain:
        result = query(
            method="domains",
            droplet_id=kwargs["dns_domain"],
            command="records",
            args={
                "type": kwargs["record_type"],
                "name": kwargs["name"],
                "data": kwargs["record_data"],
            },
            http_method="post",
        )
        return result

    return False


def destroy_dns_records(fqdn):
    """
    Deletes DNS records for the given hostname if the domain is managed with DO.
    """
    domain = ".".join(fqdn.split(".")[-2:])
    hostname = ".".join(fqdn.split(".")[:-2])
    # TODO: remove this when the todo on 754 is available
    try:
        response = query(method="domains", droplet_id=domain, command="records")
    except SaltCloudSystemExit:
        log.debug("Failed to find domains.")
        return False
    log.debug("found DNS records: %s", pprint.pformat(response))
    records = response["domain_records"]

    if records:
        record_ids = [r["id"] for r in records if r["name"].decode() == hostname]
        log.debug("deleting DNS record IDs: %s", record_ids)
        for id_ in record_ids:
            try:
                log.info("deleting DNS record %s", id_)
                ret = query(
                    method="domains",
                    droplet_id=domain,
                    command=f"records/{id_}",
                    http_method="delete",
                )
            except SaltCloudSystemExit:
                log.error(
                    "failed to delete DNS domain %s record ID %s.", domain, hostname
                )
            log.debug("DNS deletion REST call returned: %s", pprint.pformat(ret))

    return False


def show_pricing(kwargs=None, call=None):
    """
    Show pricing for a particular profile. This is only an estimate, based on
    unofficial pricing sources.

    .. versionadded:: 2015.8.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f show_pricing my-digitalocean-config profile=my-profile
    """
    profile = __opts__["profiles"].get(kwargs["profile"], {})
    if not profile:
        return {"Error": "The requested profile was not found"}

    # Make sure the profile belongs to DigitalOcean
    provider = profile.get("provider", "0:0")
    comps = provider.split(":")
    if len(comps) < 2 or comps[1] != "digitalocean":
        return {"Error": "The requested profile does not belong to DigitalOcean"}

    raw = {}
    ret = {}
    sizes = avail_sizes()
    ret["per_hour"] = decimal.Decimal(sizes[profile["size"]]["price_hourly"])

    ret["per_day"] = ret["per_hour"] * 24
    ret["per_week"] = ret["per_day"] * 7
    ret["per_month"] = decimal.Decimal(sizes[profile["size"]]["price_monthly"])
    ret["per_year"] = ret["per_week"] * 52

    if kwargs.get("raw", False):
        ret["_raw"] = raw

    return {profile["profile"]: ret}


def list_floating_ips(call=None):
    """
    Return a list of the floating ips that are on the provider

    .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f list_floating_ips my-digitalocean-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_floating_ips function must be called with "
            "-f or --function, or with the --list-floating-ips option"
        )

    fetch = True
    page = 1
    ret = {}

    while fetch:
        items = query(
            method="floating_ips",
            command="?page=" + str(page) + "&per_page=200",
        )

        for floating_ip in items["floating_ips"]:
            ret[floating_ip["ip"]] = {}
            for item in floating_ip.keys():
                ret[floating_ip["ip"]][item] = floating_ip[item]

        page += 1
        try:
            fetch = "next" in items["links"]["pages"]
        except KeyError:
            fetch = False

    return ret


def show_floating_ip(kwargs=None, call=None):
    """
    Show the details of a floating IP

    .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f show_floating_ip my-digitalocean-config floating_ip='45.55.96.47'
    """
    if call != "function":
        log.error("The show_floating_ip function must be called with -f or --function.")
        return False

    if not kwargs:
        kwargs = {}

    if "floating_ip" not in kwargs:
        log.error("A floating IP is required.")
        return False

    floating_ip = kwargs["floating_ip"]
    log.debug("Floating ip is %s", floating_ip)

    details = query(method="floating_ips", command=floating_ip)

    return details


def create_floating_ip(kwargs=None, call=None):
    """
    Create a new floating IP

    .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f create_floating_ip my-digitalocean-config region='NYC2'

        salt-cloud -f create_floating_ip my-digitalocean-config droplet_id='1234567'
    """
    if call != "function":
        log.error(
            "The create_floating_ip function must be called with -f or --function."
        )
        return False

    if not kwargs:
        kwargs = {}

    if "droplet_id" in kwargs:
        result = query(
            method="floating_ips",
            args={"droplet_id": kwargs["droplet_id"]},
            http_method="post",
        )

        return result

    elif "region" in kwargs:
        result = query(
            method="floating_ips", args={"region": kwargs["region"]}, http_method="post"
        )

        return result

    else:
        log.error("A droplet_id or region is required.")
        return False


def delete_floating_ip(kwargs=None, call=None):
    """
    Delete a floating IP

    .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f delete_floating_ip my-digitalocean-config floating_ip='45.55.96.47'
    """
    if call != "function":
        log.error(
            "The delete_floating_ip function must be called with -f or --function."
        )
        return False

    if not kwargs:
        kwargs = {}

    if "floating_ip" not in kwargs:
        log.error("A floating IP is required.")
        return False

    floating_ip = kwargs["floating_ip"]
    log.debug("Floating ip is %s", kwargs["floating_ip"])

    result = query(method="floating_ips", command=floating_ip, http_method="delete")

    return result


def assign_floating_ip(kwargs=None, call=None):
    """
    Assign a floating IP

    .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f assign_floating_ip my-digitalocean-config droplet_id=1234567 floating_ip='45.55.96.47'
    """
    if call != "function":
        log.error(
            "The assign_floating_ip function must be called with -f or --function."
        )
        return False

    if not kwargs:
        kwargs = {}

    if "floating_ip" and "droplet_id" not in kwargs:
        log.error("A floating IP and droplet_id is required.")
        return False

    result = query(
        method="floating_ips",
        command=kwargs["floating_ip"] + "/actions",
        args={"droplet_id": kwargs["droplet_id"], "type": "assign"},
        http_method="post",
    )

    return result


def unassign_floating_ip(kwargs=None, call=None):
    """
    Unassign a floating IP

    .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f unassign_floating_ip my-digitalocean-config floating_ip='45.55.96.47'
    """
    if call != "function":
        log.error(
            "The inassign_floating_ip function must be called with -f or --function."
        )
        return False

    if not kwargs:
        kwargs = {}

    if "floating_ip" not in kwargs:
        log.error("A floating IP is required.")
        return False

    result = query(
        method="floating_ips",
        command=kwargs["floating_ip"] + "/actions",
        args={"type": "unassign"},
        http_method="post",
    )

    return result


def _list_nodes(full=False, for_output=False):
    """
    Helper function to format and parse node data.
    """
    fetch = True
    page = 1
    ret = {}

    while fetch:
        items = query(method="droplets", command="?page=" + str(page) + "&per_page=200")
        for node in items["droplets"]:
            name = node["name"]
            ret[name] = {}
            if full:
                ret[name] = _get_full_output(node, for_output=for_output)
            else:
                public_ips, private_ips = _get_ips(node["networks"])
                ret[name] = {
                    "id": node["id"],
                    "image": node["image"]["name"],
                    "name": name,
                    "private_ips": private_ips,
                    "public_ips": public_ips,
                    "size": node["size_slug"],
                    "state": str(node["status"]),
                }

        page += 1
        try:
            fetch = "next" in items["links"]["pages"]
        except KeyError:
            fetch = False

    return ret


def reboot(name, call=None):
    """
    Reboot a droplet in DigitalOcean.

    .. versionadded:: 2015.8.8

    name
        The name of the droplet to restart.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a reboot droplet_name
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The reboot action must be called with -a or --action."
        )

    data = show_instance(name, call="action")
    if data.get("status") == "off":
        return {
            "success": True,
            "action": "stop",
            "status": "off",
            "msg": "Machine is already off.",
        }

    ret = query(
        droplet_id=data["id"],
        command="actions",
        args={"type": "reboot"},
        http_method="post",
    )

    return {
        "success": True,
        "action": ret["action"]["type"],
        "state": ret["action"]["status"],
    }


def start(name, call=None):
    """
    Start a droplet in DigitalOcean.

    .. versionadded:: 2015.8.8

    name
        The name of the droplet to start.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a start droplet_name
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The start action must be called with -a or --action."
        )

    data = show_instance(name, call="action")
    if data.get("status") == "active":
        return {
            "success": True,
            "action": "start",
            "status": "active",
            "msg": "Machine is already running.",
        }

    ret = query(
        droplet_id=data["id"],
        command="actions",
        args={"type": "power_on"},
        http_method="post",
    )

    return {
        "success": True,
        "action": ret["action"]["type"],
        "state": ret["action"]["status"],
    }


def stop(name, call=None):
    """
    Stop a droplet in DigitalOcean.

    .. versionadded:: 2015.8.8

    name
        The name of the droplet to stop.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop droplet_name
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    data = show_instance(name, call="action")
    if data.get("status") == "off":
        return {
            "success": True,
            "action": "stop",
            "status": "off",
            "msg": "Machine is already off.",
        }

    ret = query(
        droplet_id=data["id"],
        command="actions",
        args={"type": "shutdown"},
        http_method="post",
    )

    return {
        "success": True,
        "action": ret["action"]["type"],
        "state": ret["action"]["status"],
    }


def _get_full_output(node, for_output=False):
    """
    Helper function for _list_nodes to loop through all node information.
    Returns a dictionary containing the full information of a node.
    """
    ret = {}
    for item in node.keys():
        value = node[item]
        if value is not None and for_output:
            value = str(value)
        ret[item] = value
    return ret


def _get_ips(networks):
    """
    Helper function for list_nodes. Returns public and private ip lists based on a
    given network dictionary.
    """
    v4s = networks.get("v4")
    v6s = networks.get("v6")
    public_ips = []
    private_ips = []

    if v4s:
        for item in v4s:
            ip_type = item.get("type")
            ip_address = item.get("ip_address")
            if ip_type == "public":
                public_ips.append(ip_address)
            if ip_type == "private":
                private_ips.append(ip_address)

    if v6s:
        for item in v6s:
            ip_type = item.get("type")
            ip_address = item.get("ip_address")
            if ip_type == "public":
                public_ips.append(ip_address)
            if ip_type == "private":
                private_ips.append(ip_address)

    return public_ips, private_ips



"""
Management of Dell DRAC

The DRAC module is used to create and manage DRAC cards on Dell servers


Ensure the user damian is present

  .. code-block:: yaml

    damian:
      drac.present:
        - name: damian
        - password: secret
        - permission: login,test_alerts,clear_logs


Ensure the user damian does not exist

  .. code-block:: yaml

    damian:
      drac.absent:
        - name: damian


Ensure DRAC network is in a consistent state

  .. code-block:: yaml

    my_network:
      drac.network:
        - ip: 10.225.108.29
        - netmask: 255.255.255.224
        - gateway: 10.225.108.1

"""

import salt.exceptions
import salt.utils.path


def __virtual__():
    """
    Ensure the racadm command is installed
    """
    if salt.utils.path.which("racadm"):
        return True
    return (False, "racadm command not found")


def present(name, password, permission):
    """
    Ensure the user exists on the Dell DRAC

    name:
        The users username

    password
        The password used to authenticate

    permission
        The permissions that should be assigned to a user
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    users = __salt__["drac.list_users"]()

    if __opts__["test"]:
        if name in users:
            ret["comment"] = f"`{name}` already exists"
        else:
            ret["comment"] = f"`{name}` will be created"
            ret["changes"] = {name: "will be created"}

        return ret

    if name in users:
        ret["comment"] = f"`{name}` already exists"
    else:
        if __salt__["drac.create_user"](name, password, permission, users):
            ret["comment"] = f"`{name}` user created"
            ret["changes"] = {name: "new user created"}
        else:
            ret["comment"] = "Unable to create user"
            ret["result"] = False

    return ret


def absent(name):
    """
    Ensure a user does not exist on the Dell DRAC

    name:
        The users username
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    users = __salt__["drac.list_users"]()

    if __opts__["test"]:
        if name in users:
            ret["comment"] = f"`{name}` is set to be deleted"
            ret["changes"] = {name: "will be deleted"}
        else:
            ret["comment"] = f"`{name}` does not exist"

        return ret

    if name in users:
        if __salt__["drac.delete_user"](name, users[name]["index"]):
            ret["comment"] = f"`{name}` deleted"
            ret["changes"] = {name: "deleted"}
        else:
            ret["comment"] = "Unable to delete user"
            ret["result"] = False
    else:
        ret["comment"] = f"`{name}` does not exist"

    return ret


def network(ip, netmask, gateway):
    """
    Ensure the DRAC network settings are consistent
    """
    ret = {"name": ip, "result": True, "changes": {}, "comment": ""}

    current_network = __salt__["drac.network_info"]()
    new_network = {}

    if ip != current_network["IPv4 settings"]["IP Address"]:
        ret["changes"].update(
            {
                "IP Address": {
                    "Old": current_network["IPv4 settings"]["IP Address"],
                    "New": ip,
                }
            }
        )

    if netmask != current_network["IPv4 settings"]["Subnet Mask"]:
        ret["changes"].update(
            {
                "Netmask": {
                    "Old": current_network["IPv4 settings"]["Subnet Mask"],
                    "New": netmask,
                }
            }
        )

    if gateway != current_network["IPv4 settings"]["Gateway"]:
        ret["changes"].update(
            {
                "Gateway": {
                    "Old": current_network["IPv4 settings"]["Gateway"],
                    "New": gateway,
                }
            }
        )

    if __opts__["test"]:
        ret["result"] = None
        return ret

    if __salt__["drac.set_network"](ip, netmask, gateway):
        if not ret["changes"]:
            ret["comment"] = "Network is in the desired state"

        return ret

    ret["result"] = False
    ret["comment"] = "unable to configure network"

    return ret



"""
Setup of Python virtualenv sandboxes.

.. versionadded:: 0.17.0
"""

import logging
import os

import salt.utils.functools
import salt.utils.platform
import salt.utils.versions
import salt.version
from salt.exceptions import CommandExecutionError, CommandNotFoundError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "virtualenv"


def __virtual__():
    if "virtualenv.create" in __salt__:
        return __virtualname__
    return (False, "virtualenv module could not be loaded")


def managed(
    name,
    venv_bin=None,
    requirements=None,
    system_site_packages=False,
    distribute=False,
    use_wheel=False,
    clear=False,
    python=None,
    extra_search_dir=None,
    never_download=None,
    prompt=None,
    user=None,
    cwd=None,
    index_url=None,
    extra_index_url=None,
    pre_releases=False,
    no_deps=False,
    pip_download=None,
    pip_download_cache=None,
    pip_exists_action=None,
    pip_ignore_installed=False,
    proxy=None,
    use_vt=False,
    env_vars=None,
    no_use_wheel=False,
    pip_upgrade=False,
    pip_pkgs=None,
    pip_no_cache_dir=False,
    pip_cache_dir=None,
    process_dependency_links=False,
    no_binary=None,
    **kwargs,
):
    """
    Create a virtualenv and optionally manage it with pip

    name
        Path to the virtualenv.

    venv_bin: virtualenv
        The name (and optionally path) of the virtualenv command. This can also
        be set globally in the minion config file as ``virtualenv.venv_bin``.

    requirements: None
        Path to a pip requirements file. If the path begins with ``salt://``
        the file will be transferred from the master file server.

    use_wheel: False
        Prefer wheel archives (requires pip >= 1.4).

    python: None
        Python executable used to build the virtualenv. When Salt is installed
        from a onedir package. You will likely want to specify which python
        interperter should be used.

    user: None
        The user under which to run virtualenv and pip.

    cwd: None
        Path to the working directory where `pip install` is executed.

    no_deps: False
        Pass `--no-deps` to `pip install`.

    pip_exists_action: None
        Default action of pip when a path already exists: (s)witch, (i)gnore,
        (w)ipe, (b)ackup.

    proxy: None
        Proxy address which is passed to `pip install`.

    env_vars: None
        Set environment variables that some builds will depend on. For example,
        a Python C-module may have a Makefile that needs INCLUDE_PATH set to
        pick up a header file while compiling.

    no_use_wheel: False
        Force to not use wheel archives (requires pip>=1.4)

    no_binary
        Force to not use binary packages (requires pip >= 7.0.0)
        Accepts either :all: to disable all binary packages, :none: to empty the set,
        or a list of one or more packages

    pip_upgrade: False
        Pass `--upgrade` to `pip install`.

    pip_pkgs: None
        As an alternative to `requirements`, pass a list of pip packages that
        should be installed.

    process_dependency_links: False
        Run pip install with the --process_dependency_links flag.

        .. versionadded:: 2017.7.0

    Also accepts any kwargs that the virtualenv module will. However, some
    kwargs, such as the ``pip`` option, require ``- distribute: True``.

    .. code-block:: yaml

        /var/www/myvirtualenv.com:
          virtualenv.managed:
            - system_site_packages: False
            - requirements: salt://REQUIREMENTS.txt
            - env_vars:
                PATH_VAR: '/usr/local/bin/'

    Current versions of Salt use onedir packages and will use onedir python
    interpreter by default. If you've installed Salt via out package
    repository. You will likely want to provide the path to the interpreter
    with which you would like to be used to create the virtual environment. The
    interpreter can be specified by providing the `python` option.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if "virtualenv.create" not in __salt__:
        ret["result"] = False
        ret["comment"] = "Virtualenv was not detected on this system"
        return ret

    if salt.utils.platform.is_windows():
        venv_py = os.path.join(name, "Scripts", "python.exe")
    else:
        venv_py = os.path.join(name, "bin", "python")
    venv_exists = os.path.exists(venv_py)

    # Bail out early if the specified requirements file can't be found
    if requirements and requirements.startswith("salt://"):
        cached_requirements = __salt__["cp.is_cached"](requirements, __env__)
        if not cached_requirements:
            # It's not cached, let's cache it.
            cached_requirements = __salt__["cp.cache_file"](requirements, __env__)
        # Check if the master version has changed.
        if cached_requirements and __salt__["cp.hash_file"](
            requirements, __env__
        ) != __salt__["cp.hash_file"](cached_requirements, __env__):
            cached_requirements = __salt__["cp.cache_file"](requirements, __env__)
        if not cached_requirements:
            ret.update(
                {
                    "result": False,
                    "comment": "pip requirements file '{}' not found".format(
                        requirements
                    ),
                }
            )
            return ret
        requirements = cached_requirements

    # If it already exists, grab the version for posterity
    if venv_exists and clear:
        ret["changes"]["cleared_packages"] = __salt__["pip.freeze"](bin_env=name)
        ret["changes"]["old"] = __salt__["cmd.run_stderr"](f"{venv_py} -V").strip("\n")

    # Create (or clear) the virtualenv
    if __opts__["test"]:
        if venv_exists and clear:
            ret["result"] = None
            ret["comment"] = f"Virtualenv {name} is set to be cleared"
            return ret
        if venv_exists and not clear:
            ret["comment"] = f"Virtualenv {name} is already created"
            return ret
        ret["result"] = None
        ret["comment"] = f"Virtualenv {name} is set to be created"
        return ret

    if not venv_exists or (venv_exists and clear):
        try:
            venv_ret = __salt__["virtualenv.create"](
                name,
                venv_bin=venv_bin,
                system_site_packages=system_site_packages,
                distribute=distribute,
                clear=clear,
                python=python,
                extra_search_dir=extra_search_dir,
                never_download=never_download,
                prompt=prompt,
                user=user,
                use_vt=use_vt,
                **kwargs,
            )
        except CommandNotFoundError as err:
            ret["result"] = False
            ret["comment"] = f"Failed to create virtualenv: {err}"
            return ret

        if venv_ret["retcode"] != 0:
            ret["result"] = False
            ret["comment"] = venv_ret["stdout"] + venv_ret["stderr"]
            return ret

        ret["result"] = True
        ret["changes"]["new"] = __salt__["cmd.run_stderr"](f"{venv_py} -V").strip("\n")

        if clear:
            ret["comment"] = "Cleared existing virtualenv"
        else:
            ret["comment"] = "Created new virtualenv"

    elif venv_exists:
        ret["comment"] = "virtualenv exists"

    # Check that the pip binary supports the 'use_wheel' option
    if use_wheel:
        min_version = "1.4"
        max_version = "9.0.3"
        cur_version = __salt__["pip.version"](bin_env=name)
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        too_high = salt.utils.versions.compare(
            ver1=cur_version, oper=">", ver2=max_version
        )
        if too_low or too_high:
            ret["result"] = False
            ret["comment"] = (
                "The 'use_wheel' option is only supported in "
                "pip between {} and {}. The version of pip detected "
                "was {}.".format(min_version, max_version, cur_version)
            )
            return ret

    # Check that the pip binary supports the 'no_use_wheel' option
    if no_use_wheel:
        min_version = "1.4"
        max_version = "9.0.3"
        cur_version = __salt__["pip.version"](bin_env=name)
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        too_high = salt.utils.versions.compare(
            ver1=cur_version, oper=">", ver2=max_version
        )
        if too_low or too_high:
            ret["result"] = False
            ret["comment"] = (
                "The 'no_use_wheel' option is only supported in "
                "pip between {} and {}. The version of pip detected "
                "was {}.".format(min_version, max_version, cur_version)
            )
            return ret

    # Check that the pip binary supports the 'no_binary' option
    if no_binary:
        min_version = "7.0.0"
        cur_version = __salt__["pip.version"](bin_env=name)
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        if too_low:
            ret["result"] = False
            ret["comment"] = (
                "The 'no_binary' option is only supported in "
                "pip {} and newer. The version of pip detected "
                "was {}.".format(min_version, cur_version)
            )
            return ret

    # Populate the venv via a requirements file
    if requirements or pip_pkgs:
        try:
            before = set(__salt__["pip.freeze"](bin_env=name, user=user, use_vt=use_vt))
        except CommandExecutionError as exc:
            ret["result"] = False
            ret["comment"] = exc.strerror
            return ret

        if requirements:

            if isinstance(requirements, str):
                req_canary = requirements.split(",")[0]
            elif isinstance(requirements, list):
                req_canary = requirements[0]
            else:
                raise TypeError("pip requirements must be either a string or a list")

            if req_canary != os.path.abspath(req_canary):
                cwd = os.path.dirname(os.path.abspath(req_canary))

        pip_ret = __salt__["pip.install"](
            pkgs=pip_pkgs,
            requirements=requirements,
            process_dependency_links=process_dependency_links,
            bin_env=name,
            use_wheel=use_wheel,
            no_use_wheel=no_use_wheel,
            no_binary=no_binary,
            user=user,
            cwd=cwd,
            index_url=index_url,
            extra_index_url=extra_index_url,
            download=pip_download,
            download_cache=pip_download_cache,
            pre_releases=pre_releases,
            exists_action=pip_exists_action,
            ignore_installed=pip_ignore_installed,
            upgrade=pip_upgrade,
            no_deps=no_deps,
            proxy=proxy,
            use_vt=use_vt,
            env_vars=env_vars,
            no_cache_dir=pip_no_cache_dir,
            cache_dir=pip_cache_dir,
            **kwargs,
        )
        ret["result"] &= pip_ret["retcode"] == 0
        if pip_ret["retcode"] > 0:
            ret["comment"] = "{}\n{}\n{}".format(
                ret["comment"], pip_ret["stdout"], pip_ret["stderr"]
            )

        after = set(__salt__["pip.freeze"](bin_env=name))

        new = list(after - before)
        old = list(before - after)

        if new or old:
            ret["changes"]["packages"] = {
                "new": new if new else "",
                "old": old if old else "",
            }
    return ret


manage = salt.utils.functools.alias_function(managed, "manage")
